# coding=utf-8
import json
import os
from unittest import mock

import pytest
import responses
from github3.pulls import ShortPullRequest
from github3.repos.repo import Repository

from cumulusci.core.exceptions import CumulusCIException
from cumulusci.core.github import get_github_api
from cumulusci.tasks.github.tests.util_github_api import GithubApiTestMixin
from cumulusci.tasks.release_notes.generator import (
    BaseReleaseNotesGenerator,
    DirectoryReleaseNotesGenerator,
    StaticReleaseNotesGenerator,
    render_empty_pr_section,
)
from cumulusci.tasks.release_notes.parser import BaseChangeNotesParser
from cumulusci.tasks.release_notes.tests.utils import MockUtil
from cumulusci.tests.util import create_project_config
from cumulusci.vcs.github.release_notes.generator import (
    GithubReleaseNotesGenerator,
    ParentPullRequestNotesGenerator,
)

__location__ = os.path.split(os.path.realpath(__file__))[0]

PARSER_CONFIG = [
    {
        "class_path": "cumulusci.vcs.github.release_notes.parser.GithubLinesParser",
        "title": "Critical Changes",
    },
    {
        "class_path": "cumulusci.vcs.github.release_notes.parser.GithubLinesParser",
        "title": "Changes",
    },
    {
        "class_path": "cumulusci.vcs.github.release_notes.parser.GithubIssuesParser",
        "title": "Issues Closed",
    },
    {"class_path": None},
]


class DummyParser(BaseChangeNotesParser):
    def parse(self, change_note):
        pass

    def _render(self):
        return "dummy parser output"


class TestBaseReleaseNotesGenerator:
    def test_render_no_parsers(self):
        release_notes = BaseReleaseNotesGenerator()
        content = release_notes.render()
        assert content == ""

    def test_render_dummy_parsers(self):
        release_notes = BaseReleaseNotesGenerator()
        release_notes.parsers.append(DummyParser("Dummy 1"))
        release_notes.parsers.append(DummyParser("Dummy 2"))
        expected = (
            "# Dummy 1\r\n\r\ndummy parser output\r\n\r\n"
            + "# Dummy 2\r\n\r\ndummy parser output"
        )
        assert release_notes.render() == expected


class TestStaticReleaseNotesGenerator:
    def test_init_parser(self):
        release_notes = StaticReleaseNotesGenerator([])
        assert len(release_notes.parsers) == 3


class TestDirectoryReleaseNotesGenerator:
    def test_init_parser(self):
        release_notes = DirectoryReleaseNotesGenerator("change_notes")
        assert len(release_notes.parsers) == 3

    def test_full_content(self):
        change_notes_dir = os.path.join(__location__, "change_notes", "full")
        release_notes = DirectoryReleaseNotesGenerator(change_notes_dir)

        content = release_notes()
        expected = "# Critical Changes\r\n\r\n* This will break everything!\r\n\r\n# Changes\r\n\r\nHere's something I did. It was really cool\r\nOh yeah I did something else too!\r\n\r\n# Issues Closed\r\n\r\n#2345\r\n#6236"
        print(expected)
        print("-------------------------------------")
        print(content)

        assert content == expected


class TestGithubReleaseNotesGenerator(GithubApiTestMixin):
    def setup_method(self):
        self.current_tag = "prod/1.4"
        self.last_tag = "prod/1.3"
        self.github_info = {
            "github_owner": "TestOwner",
            "github_repo": "TestRepo",
            "github_username": "TestUser",
            "github_password": "TestPass",
        }
        self.gh = get_github_api("TestUser", "TestPass")
        self.mock_util = MockUtil("TestOwner", "TestRepo")

    @responses.activate
    def test_init_without_last_tag(self):
        github_info = self.github_info.copy()
        self.mock_util.mock_get_repo()
        generator = GithubReleaseNotesGenerator(
            self.gh, github_info, PARSER_CONFIG, self.current_tag, version_id="04t"
        )
        assert generator.github_info == github_info
        assert generator.current_tag == self.current_tag
        assert generator.last_tag is None
        assert generator.change_notes.current_tag == self.current_tag
        assert generator.change_notes._last_tag is None
        assert generator.version_id == "04t"

    @responses.activate
    def test_init_with_last_tag(self):
        github_info = self.github_info.copy()
        self.mock_util.mock_get_repo()
        generator = GithubReleaseNotesGenerator(
            self.gh, github_info, PARSER_CONFIG, self.current_tag, self.last_tag
        )
        assert generator.github_info == github_info
        assert generator.current_tag == self.current_tag
        assert generator.last_tag == self.last_tag
        assert generator.change_notes.current_tag == self.current_tag
        assert generator.change_notes._last_tag == self.last_tag

    @responses.activate
    def test_render_empty_pr_section(self):
        self.mock_util.mock_get_repo()
        self.mock_util.mock_pull_request(1, body="# Changes\r\n\r\nfoo")
        self.mock_util.mock_pull_request(2, body="# Changes\r\n\r\nbar")
        generator = self._create_generator()
        repo = generator.get_repo()
        pr1 = repo.pull_request(1)
        pr2 = repo.pull_request(2)
        generator.empty_change_notes.extend([pr1, pr2])
        content = render_empty_pr_section(generator.empty_change_notes)
        assert 3 == len(content)
        assert content[0] == "\n# Pull requests with no release notes"
        assert (
            "\n* {} [[PR{}]({})]".format(pr1.title, pr1.number, pr1.html_url)
            == content[1]
        )
        assert (
            "\n* {} [[PR{}]({})]".format(pr2.title, pr2.number, pr2.html_url)
            == content[2]
        )

    @responses.activate
    def test_update_content_with_empty_release_body(self):
        self.mock_util.mock_get_repo()
        self.mock_util.mock_pull_request(88, body="Just a small note.")
        self.mock_util.mock_pull_request(89, body="")
        generator = self._create_generator()
        repo = generator.get_repo()
        pr1 = repo.pull_request(88)
        pr2 = repo.pull_request(89)
        generator.include_empty_pull_requests = True
        generator.empty_change_notes = [pr1, pr2]
        release = mock.Mock(body=None)
        content = generator._update_release_content(release, "new content")

        split_content = content.split("\r\n")
        assert 4 == len(split_content)
        assert split_content[0] == "new content"
        assert split_content[1] == "\n# Pull requests with no release notes"
        assert (
            "\n* Pull Request #{0} [[PR{0}]({1})]".format(pr1.number, pr1.html_url)
            == split_content[2]
        )
        assert (
            "\n* Pull Request #{0} [[PR{0}]({1})]".format(pr2.number, pr2.html_url)
            == split_content[3]
        )

    @responses.activate
    def test_detect_empty_change_note(self):
        self.mock_util.mock_get_repo()
        self.mock_util.mock_pull_request(1, body="# Changes\r\n\r\nfoo")
        self.mock_util.mock_pull_request(2, body="Nothing under headers we track")
        self.mock_util.mock_pull_request(3, body="")
        generator = self._create_generator()
        repo = generator.get_repo()
        pr1 = repo.pull_request(1)
        pr2 = repo.pull_request(2)
        pr3 = repo.pull_request(3)

        generator._parse_change_note(pr1)
        generator._parse_change_note(pr2)
        generator._parse_change_note(pr3)

        # PR1 is "non-empty" second two are "empty"
        assert 2 == len(generator.empty_change_notes)
        assert 2 == generator.empty_change_notes[0].number
        assert 3 == generator.empty_change_notes[1].number

    def _create_generator(self):
        generator = GithubReleaseNotesGenerator(
            self.gh, self.github_info.copy(), PARSER_CONFIG, self.current_tag
        )
        return generator


class TestPublishingGithubReleaseNotesGenerator(GithubApiTestMixin):
    def setup_method(self):
        self.init_github()
        self.github_info = {
            "github_owner": "TestOwner",
            "github_repo": "TestRepo",
            "github_username": "TestUser",
            "github_password": "TestPass",
            "default_branch": "main",
        }
        self.gh = get_github_api("TestUser", "TestPass")
        self.mock_util = MockUtil("TestOwner", "TestRepo")

    @responses.activate
    def test_publish_update_unicode(self):
        tag = "prod/1.4"
        note = "“Unicode quotes”"
        expected_release_body = "# Changes\r\n\r\n{}".format(note)
        # mock GitHub API responses
        self.mock_util.mock_get_repo()
        # create generator instance
        generator = self._create_generator(tag)
        # inject content into Changes parser
        generator.parsers[1].content.append(note)
        # render content
        content = generator.render()
        # verify
        assert len(responses.calls._calls) == 1
        assert content == expected_release_body

    @responses.activate
    def test_publish_update_no_body(self):
        tag = "prod/1.4"
        expected_release_body = "# Changes\r\n\r\nfoo"
        # mock GitHub API responses
        self.mock_util.mock_get_repo()
        # create generator
        generator = self._create_generator(tag)
        # inject content into Changes parser
        generator.parsers[1].content.append("foo")
        # render content
        content = generator.render()
        # verify
        assert len(responses.calls._calls) == 1
        assert content == expected_release_body

    @responses.activate
    def test_publish_update_content_before(self):
        tag = "prod/1.4"
        expected_release_body = "foo\r\n# Changes\r\n\r\nbaz"
        # mock GitHub API responses
        self.mock_util.mock_get_repo()
        self.mock_util.mock_get_release(tag=tag, body="foo\n# Changes\nbar")
        # create generator
        generator = self._create_generator(tag)
        # inject content into parser
        generator.parsers[1].content.append("baz")
        # render and update content
        content = generator.render()
        release = generator._get_release()
        content = generator._update_release_content(release, content)
        # verify
        assert len(responses.calls._calls) == 3
        assert content == expected_release_body

    @responses.activate
    def test_publish_update_content_after(self):
        tag = "prod/1.4"
        expected_release_body = "# Changes\r\n\r\nbaz\r\n\r\n# Foo\r\nfoo"
        # mock GitHub API responses
        self.mock_util.mock_get_repo()
        self.mock_util.mock_get_release(tag=tag, body="# Changes\nbar\n# Foo\nfoo")
        # create generator
        generator = self._create_generator(tag)
        # inject content into parser
        generator.parsers[1].content.append("baz")
        # render and update content
        content = generator.render()
        release = generator._get_release()
        content = generator._update_release_content(release, content)
        # verify
        assert len(responses.calls._calls) == 3
        assert content == expected_release_body

    @responses.activate
    def test_publish_update_content_before_and_after(self):
        tag = "prod/1.4"
        expected_release_body = "foo\r\n# Changes\r\n\r\nbaz\r\n\r\n# Foo\r\nfoo"
        # mock GitHub API responses
        self.mock_util.mock_get_repo()
        self.mock_util.mock_get_release(tag=tag, body="foo\n# Changes\nbar\n# Foo\nfoo")
        # create generator
        generator = self._create_generator(tag)
        # inject content into parser
        generator.parsers[1].content.append("baz")
        # render and update content
        content = generator.render()
        release = generator._get_release()
        content = generator._update_release_content(release, content)
        # verify
        assert len(responses.calls._calls) == 3
        assert content == expected_release_body

    @responses.activate
    def test_publish_update_content_between(self):
        tag = "prod/1.4"
        expected_release_body = (
            "# Critical Changes\r\n\r\nfaz\r\n\r\n"
            "# Foo\r\nfoo\r\n# Changes\r\n\r\nfiz"
        )
        # mock GitHub API responses
        self.mock_util.mock_get_repo()
        self.mock_util.mock_get_release(
            tag=tag, body="# Critical Changes\nbar\n# Foo\nfoo\n# Changes\nbiz"
        )
        # create generator
        generator = self._create_generator(tag)
        # inject content into parser
        generator.parsers[0].content.append("faz")
        generator.parsers[1].content.append("fiz")
        # render and update content
        content = generator.render()
        release = generator._get_release()
        content = generator._update_release_content(release, content)
        # verify
        assert len(responses.calls._calls) == 3
        assert content == expected_release_body

    @responses.activate
    def test_publish_update_content_before_after_and_between(self):
        tag = "prod/1.4"
        expected_release_body = (
            "goo\r\n# Critical Changes\r\n\r\nfaz\r\n\r\n"
            "# Foo\r\nfoo\r\n# Changes\r\n\r\nfiz\r\n\r\n# Zoo\r\nzoo"
        )
        # mock GitHub API responses
        self.mock_util.mock_get_repo()
        self.mock_util.mock_get_release(
            tag=tag,
            body=(
                "goo\n# Critical Changes\nbar\n"
                "# Foo\nfoo\n# Changes\nbiz\n# Zoo\nzoo"
            ),
        )
        # create generator
        generator = self._create_generator(tag)
        # inject content into parser
        generator.parsers[0].content.append("faz")
        generator.parsers[1].content.append("fiz")
        # render and update content
        content = generator.render()
        release = generator._get_release()
        content = generator._update_release_content(release, content)
        # verify
        assert len(responses.calls._calls) == 3
        assert content == expected_release_body

    @responses.activate
    def test_publish_update_content_missing(self):
        tag = "prod/1.4"
        expected_release_body = "foo\r\n# Changes\r\n\r\nbaz\r\n"
        # mock GitHub API responses
        self.mock_util.mock_get_repo()
        self.mock_util.mock_get_release(tag=tag, body="foo")
        # create generator
        generator = self._create_generator(tag)
        # inject content into parser
        generator.parsers[1].content.append("baz")
        # render and update content
        content = generator.render()
        release = generator._get_release()
        content = generator._update_release_content(release, content)
        # verify
        assert len(responses.calls._calls) == 3
        assert content == expected_release_body

    def _create_generator(self, current_tag, last_tag=None):
        generator = GithubReleaseNotesGenerator(
            self.gh, self.github_info.copy(), PARSER_CONFIG, current_tag, last_tag
        )
        return generator

    @mock.patch(
        "cumulusci.tasks.release_notes.generator.BaseReleaseNotesGenerator.__call__"
    )
    @responses.activate
    def test_call(self, base_generator):
        self.mock_util.mock_get_repo()
        generator = self._create_generator("prod/1.0")
        generator.do_publish = True
        release = mock.Mock()
        generator._get_release = mock.Mock(return_value=release)
        generator._update_release_content = mock.Mock(
            return_value=mock.sentinel.content
        )
        result = generator()
        assert mock.sentinel.content is result
        base_generator.assert_called_once()
        release.edit.assert_called_once()

    @responses.activate
    def test_call__no_release(self):
        self.mock_util.mock_get_repo()
        generator = self._create_generator("prod/1.0")
        responses.add(
            method=responses.GET,
            url="{}/releases/tags/prod/1.0".format(self.mock_util.repo_url),
            status=404,
        )
        with pytest.raises(CumulusCIException):
            generator()


class TestParentPullRequestNotesGenerator(GithubApiTestMixin):
    @pytest.fixture
    def mock_util(self):
        return MockUtil("TestOwner", "TestRepo")

    @pytest.fixture
    def repo(self, gh_api):
        repo_json = self._get_expected_repo("TestOwner", "TestRepo")
        return Repository(repo_json, gh_api)

    @pytest.fixture
    def generator(self, gh_api):
        repo_json = GithubApiTestMixin()._get_expected_repo("TestOwner", "TestRepo")
        repo = Repository(repo_json, gh_api)
        return ParentPullRequestNotesGenerator(gh_api, repo, create_project_config())

    def test_init_parsers(self, generator):
        assert 7 == len(generator.parsers)
        assert generator.parsers[-1]._in_section

    @responses.activate
    def test_aggregate_child_change_notes(self, generator, mock_util, gh_api):
        def request_intercept(request):
            """Assert that the body is correct"""
            body = json.loads(request.body)["body"]
            assert (
                "# Critical Changes\r\n\r\n* Everything works now. [[PR2](https://github.com/TestOwner/TestRepo/pulls/2)]"
                in body
            )
            assert (
                "# Changes\r\n\r\n* Now, more code! [[PR1](https://github.com/TestOwner/TestRepo/pulls/1)]"
                in body
            )
            assert "# Notes From Child PRs\r\n\r\n* Dev note 1 [[PR1](https://github.com/TestOwner/TestRepo/pulls/1)]\r\n\r\n* Dev note 2 [[PR2](https://github.com/TestOwner/TestRepo/pulls/2)]"
            assert (
                "# Pull requests with no release notes\r\n\n* Pull Request #4 [[PR4](https://github.com/TestOwner/TestRepo/pulls/4)]"
                in body
            )
            assert "Should not be in body" not in body
            return (200, {}, json.dumps(self._get_expected_pull_request(3, 3, body)))

        self.init_github()
        dev_note1 = "* Dev note 1"
        dev_note2 = "* Dev note 2"
        change = "# Changes\r\n\r\n* Now, more code!"
        critical_change = "# Critical Changes\r\n\r\n* Everything works now."

        pr1_json = self._get_expected_pull_request(1, 1, dev_note1 + "\r\n" + change)
        pr2_json = self._get_expected_pull_request(
            2, 2, dev_note2 + "\r\n" + critical_change
        )
        pr3_json = self._get_expected_pull_request(4, 4, None)
        pr3_json["body"] = ""

        pr4_json = self._get_expected_pull_request(5, 5, "Should not be in body")
        # simulate merge from main back into parent
        pr4_json["head"]["ref"] = "main"

        mock_util.mock_pulls(pulls=[pr1_json, pr2_json, pr3_json, pr4_json])

        pr_json = self._get_expected_pull_request(3, 3, "Body of Parent PR")
        parent_pr = ShortPullRequest(pr_json, gh_api)
        parent_pr.head.label = "repo:some-other-branch"

        responses.add_callback(
            responses.PATCH,
            "https://github.com/TestOwner/TestRepo/pulls/3",  # TODO: again, no '.api' needed in this endpoint?
            callback=request_intercept,
            content_type="application/json",
        )
        generator.aggregate_child_change_notes(parent_pr)

    @mock.patch(
        "cumulusci.vcs.github.release_notes.generator.get_pull_requests_with_base_branch"
    )
    def test_aggregate_child_change_notes__update_fails(
        self, get_pull, generator, mock_util, gh_api
    ):
        self.init_github()
        parent_body = "Body of Parent PR"
        pr_json = self._get_expected_pull_request(3, 3, parent_body)
        parent_pr = ShortPullRequest(pr_json, gh_api)
        parent_pr.merged_at = "Yesterday"
        parent_pr.head.label = "repo:some-other-branch"
        parent_pr.update = mock.Mock(return_value=False)

        get_pull.return_value = [parent_pr]
        with pytest.raises(CumulusCIException):
            generator.aggregate_child_change_notes(parent_pr)
        parent_pr.update.assert_called_once()

    @responses.activate
    def test_aggregate_child_change_notes__empty_change_note(
        self, generator, mock_util, gh_api
    ):
        self.init_github()
        mock_util.mock_pulls()  # no change notes returned

        parent_pr = ShortPullRequest(
            self._get_expected_pull_request(3, 3, "Body"), gh_api
        )
        parent_pr.head.label = "repo:branch"

        generator.aggregate_child_change_notes(parent_pr)
        assert 0 == len(generator.change_notes)


class TestMigrationGenerator(GithubApiTestMixin):
    def setup_method(self):
        self.current_tag = "prod/1.4"
        self.last_tag = "prod/1.3"
        self.github_info = {
            "github_owner": "TestOwner",
            "github_repo": "TestRepo",
            "github_username": "TestUser",
            "github_password": "TestPass",
        }
        self.gh = get_github_api("TestUser", "TestPass")
        self.mock_util = MockUtil("TestOwner", "TestRepo")

    @responses.activate
    def test_migration(self):
        from cumulusci.tasks.release_notes.generator import (
            GithubReleaseNotesGenerator,
            ParentPullRequestNotesGenerator,
        )
        from cumulusci.utils.deprecation import ClassMovedWarning

        github_info = self.github_info.copy()
        self.mock_util.mock_get_repo()
        with pytest.warns(
            ClassMovedWarning, match="cumulusci.vcs.github.release_notes.generator"
        ):
            GithubReleaseNotesGenerator(
                self.gh, github_info, PARSER_CONFIG, self.current_tag, version_id="04t"
            )

        with pytest.warns(
            ClassMovedWarning, match="cumulusci.vcs.github.release_notes.generator"
        ):
            repo_json = GithubApiTestMixin()._get_expected_repo("TestOwner", "TestRepo")
            repo = Repository(repo_json, self.gh)
            ParentPullRequestNotesGenerator(self.gh, repo, create_project_config())

        assert (
            GithubReleaseNotesGenerator.__module__
            == "cumulusci.tasks.release_notes.generator"
        )
        assert GithubReleaseNotesGenerator.__name__ == "GithubReleaseNotesGenerator"
        assert (
            ParentPullRequestNotesGenerator.__module__
            == "cumulusci.tasks.release_notes.generator"
        )
        assert (
            ParentPullRequestNotesGenerator.__name__
            == "ParentPullRequestNotesGenerator"
        )



import re

import pytest
import requests
import responses

from cumulusci.tasks.marketing_cloud.get_user_info import GetUserInfoTask
from cumulusci.tasks.marketing_cloud.mc_constants import MC_API_VERSION


@pytest.fixture
def get_user_info_task(create_task, mc_project_config):
    return create_task(GetUserInfoTask, {}, project_config=mc_project_config)


@pytest.fixture
def user_info_payload():
    return {
        "exp": 1527771992,
        "iss": "https://mc.exacttarget.com",
        "user": {
            "sub": "10654321",
            "name": "Auth_user_name",
            "preferred_username": "Auth_user_preferred_username",
            "email": "example@example.com",
            "locale": "en-GB",
            "zoneinfo": "Europe/London",
            "timezone": {
                "longName": "(GMT) Dublin, Edinburgh, Lisbon, London *",
                "shortName": "GMT+0",
                "offset": 0,
                "dst": True,
            },
        },
        "organization": {
            "member_id": 10123456,
            "enterprise_id": 10123456,
            "enterprise_name": "Auth_enterprise_name",
            "account_type": "enterprise",
            "stack_key": "S1",
            "region": "NA1",
            "locale": "en-US",
            "zoneinfo": "America/Los_Angeles",
            "timezone": {
                "longName": "(GMT-08:00) Pacific Time (US & Canada) *",
                "shortName": "GMT-8",
                "offset": -8,
                "dst": True,
            },
        },
        "rest": {
            "rest_instance_url": "https://mc563885gzs27c5t9-63k636ttgm.rest.marketingcloudapis.com",
            "soap_instance_url": "https://mc563885gzs27c5t9-63k636ttgm.soap.marketingcloudapis.com",
        },
        "application": {
            "id": "1a23b4cd-5e66-789f-0g1h-2i3a6efb6d80",
            "name": "auth_application_name",
            "redirectUrl": [
                "https://example.example.com*",
                "https://example.com/oauth-authorize",
                "https://example.example.com/***********",
            ],
            "appScopes": [
                "openid",
                "offline",
                "email_read",
                "email_send",
                "email_write",
            ],
        },
        "permissions": [
            {
                "objectTypeName": "Email",
                "operationName": "Update",
                "name": "Update",
                "id": 123,
            }
        ],
    }


@responses.activate
def test_get_user_info__success(get_user_info_task, user_info_payload):
    responses.add(
        "POST",
        "https://tssd.auth.marketingcloudapis.com/v2/token",
        json={"access_token": "ACCESS_TOKEN"},
    )
    responses.add(
        "GET",
        f"https://TSSD.auth.marketingcloudapis.com/{MC_API_VERSION}/userinfo",
        json=user_info_payload,
    )

    get_user_info_task()

    expected_payload = user_info_payload
    del expected_payload["rest"]
    del expected_payload["application"]
    del expected_payload["permissions"]

    assert get_user_info_task.return_values == expected_payload


@responses.activate
def test_get_user_info__HTTPError(get_user_info_task, user_info_payload):
    responses.add(
        "POST",
        "https://tssd.auth.marketingcloudapis.com/v2/token",
        json={"access_token": "ACCESS_TOKEN"},
    )
    responses.add(
        "GET",
        f"https://TSSD.auth.marketingcloudapis.com/{MC_API_VERSION}/userinfo",
        status=400,
        body="bad request",
    )
    with pytest.raises(
        requests.exceptions.HTTPError,
        match=re.escape(
            "400 Client Error: Bad Request for url: https://tssd.auth.marketingcloudapis.com/v2/userinfo"
        ),
    ):
        get_user_info_task()


def test_sanitize_payload(get_user_info_task):
    payload = {
        "foo": "foo-stuff",
        "bar": "bar-stuff",
        "rest": "rest-stuff",
        "application": "application-stuff",
        "permissions": "permission-stuff",
    }
    start_num_keys = len(payload.keys())
    sanitized_payload = get_user_info_task._sanitize_payload(payload)

    assert start_num_keys - 3 == len(sanitized_payload.keys())
    assert "rest" not in sanitized_payload
    assert "appliation" not in sanitized_payload
    assert "permissions" not in sanitized_payload



import json
import os
from datetime import datetime
from ssl import SSLCertVerificationError
from unittest import mock

import pytest
import requests
import responses
from github3 import GitHub, GitHubEnterprise
from github3.exceptions import (
    AuthenticationFailed,
    ConnectionError,
    ForbiddenError,
    ResponseError,
    TransportError,
)
from github3.pulls import ShortPullRequest
from github3.repos.repo import Repository
from github3.session import AppInstallationTokenAuth
from requests.exceptions import RequestException, RetryError, SSLError
from requests.models import Response

import cumulusci
from cumulusci.core import github
from cumulusci.core.config import BaseProjectConfig, ServiceConfig, UniversalConfig
from cumulusci.core.exceptions import (
    DependencyLookupError,
    GithubApiError,
    GithubApiNotFoundError,
    GithubException,
    ServiceNotConfigured,
)
from cumulusci.core.github import (
    SELF_SIGNED_WARNING,
    SSO_WARNING,
    UNAUTHORIZED_WARNING,
    _determine_github_client,
    add_labels_to_pull_request,
    catch_common_github_auth_errors,
    check_github_sso_auth,
    create_gist,
    create_pull_request,
    format_github3_exception,
    get_auth_from_service,
    get_commit,
    get_github_api,
    get_github_api_for_repo,
    get_latest_prerelease,
    get_oauth_device_flow_token,
    get_oauth_scopes,
    get_pull_requests_by_commit,
    get_pull_requests_by_head,
    get_pull_requests_with_base_branch,
    get_ref_for_tag,
    get_sso_disabled_orgs,
    get_tag_by_name,
    get_version_id_from_tag,
    is_label_on_pull_request,
    is_pull_request_merged,
    markdown_link_to_pr,
    request_url_from_exc,
    validate_gh_enterprise,
    validate_service,
    warn_oauth_restricted,
)
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.tasks.github.tests.util_github_api import GithubApiTestMixin
from cumulusci.tasks.release_notes.tests.utils import MockUtil


class TestGithub(GithubApiTestMixin):
    @classmethod
    def teardown_method(cls):
        # clear cached repo -> installation mapping
        github.INSTALLATIONS.clear()

    @pytest.fixture
    def mock_util(self):
        return MockUtil("TestOwner", "TestRepo")

    @pytest.fixture
    def repo(self, gh_api):
        repo_json = self._get_expected_repo("TestOwner", "TestRepo")
        return Repository(repo_json, gh_api)

    @pytest.fixture
    def keychain_enterprise(self):
        runtime = mock.Mock()
        runtime.project_config = BaseProjectConfig(UniversalConfig(), config={})
        runtime.keychain = BaseProjectKeychain(runtime.project_config, None)
        runtime.keychain.set_service(
            "github_enterprise",
            "ent",
            ServiceConfig(
                {
                    "username": "testusername",
                    "email": "test@domain.com",
                    "token": "ATOKEN",
                    "server_domain": "git.enterprise.domain.com",
                }
            ),
        )
        return runtime.keychain

    def test_github_api_retries__escape_hatch_SSL_error(self):
        gh = get_github_api("TestUser", "TestPass")
        adapter = gh.session.get_adapter("http://")

        assert 0.3 == adapter.max_retries.backoff_factor
        assert 502 in adapter.max_retries.status_forcelist

        with mock.patch(
            "urllib3.connectionpool.HTTPConnectionPool._make_request"
        ) as _make_request:
            _make_request.side_effect = SSLCertVerificationError(
                "SSL: CERTIFICATE_VERIFY_FAILED] certificate verify failed: self signed certificate in certificate chain"
            )

            with pytest.raises(
                ConnectionError,
                match="(CERTIFICATE_VERIFY_FAILED)",
            ):
                gh.octocat("meow")

            assert 1 == _make_request.call_count

    def test_github_api_retries(self):
        gh = get_github_api("TestUser", "TestPass")
        adapter = gh.session.get_adapter("http://")

        assert 0.3 == adapter.max_retries.backoff_factor
        assert 502 in adapter.max_retries.status_forcelist

        user_json = {
            "login": "TestUser",
            "id": 123456,
            "type": "User",
            "url": "https://api.github.com/user",
            "avatar_url": "https://avatars.githubusercontent.com/u/123456?v=4",
            "events_url": "https://api.github.com/users/TestUser/events{/privacy}",
            "followers_url": "https://api.github.com/users/TestUser/followers",
            "following_url": "https://api.github.com/users/TestUser/following{/other_user}",
            "gists_url": "https://api.github.com/users/TestUser/gists{/gist_id}",
            "gravatar_id": "",
            "html_url": "https://github.com/TestUser",
            "organizations_url": "https://api.github.com/users/TestUser/orgs",
            "received_events_url": "https://api.github.com/users/TestUser/received_events",
            "repos_url": "https://api.github.com/users/TestUser/repos",
            "site_admin": False,
            "starred_url": "https://api.github.com/users/TestUser/starred{/owner}{/repo}",
            "subscriptions_url": "https://api.github.com/users/TestUser/subscriptions",
            "bio": "Test bio",
            "blog": "https://testuserblog.com",
            "company": "Test Company",
            "created_at": "2020-01-01T00:00:00Z",
            "email": "testuser@example.com",
            "followers": 10,
            "following": 5,
            "hireable": True,
            "location": "Test City",
            "name": "Test User",
            "public_gists": 2,
            "public_repos": 3,
            "updated_at": "2023-01-01T00:00:00Z",
        }

        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, "https://api.github.com/user", status=503)
            rsps.add(
                responses.GET, "https://api.github.com/user", json=user_json, status=200
            )

            gh.me()
            assert len(rsps.calls) == 2

    @responses.activate
    @mock.patch("github3.apps.create_token")
    def test_get_github_api_for_repo(self, create_token):
        create_token.return_value = "ATOKEN"
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/installation",
            json={
                "id": 1,
                "access_tokens_url": "",
                "account": "",
                "app_id": "",
                "created_at": "",
                "events": "",
                "html_url": "",
                "permissions": "",
                "repositories_url": "",
                "repository_selection": "",
                "single_file_name": "",
                "target_id": "",
                "target_type": "",
                "updated_at": "",
            },
        )
        responses.add(
            "POST",
            "https://api.github.com/app/installations/1/access_tokens",
            status=201,
            json={"token": "ITOKEN", "expires_at": datetime.now().isoformat()},
        )

        with mock.patch.dict(
            os.environ, {"GITHUB_APP_KEY": "bogus", "GITHUB_APP_ID": "1234"}
        ):
            gh = get_github_api_for_repo(None, "https://github.com/TestOwner/TestRepo/")
            assert isinstance(gh.session.auth, AppInstallationTokenAuth)

    @responses.activate
    @mock.patch("github3.apps.create_token")
    def test_get_github_api_for_repo__not_installed(self, create_token):
        create_token.return_value = "ATOKEN"
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/installation",
            status=404,
        )
        with mock.patch.dict(
            os.environ, {"GITHUB_APP_KEY": "bogus", "GITHUB_APP_ID": "1234"}
        ):
            with pytest.raises(GithubException):
                get_github_api_for_repo(None, "https://github.com/TestOwner/TestRepo/")

    @responses.activate
    @mock.patch("cumulusci.core.github.GitHub")
    def test_get_github_api_for_repo__token(self, GitHub):
        with mock.patch.dict(os.environ, {"GITHUB_TOKEN": "token"}):
            gh = get_github_api_for_repo(None, "https://github.com/TestOwner/TestRepo/")
        gh.login.assert_called_once_with(token="token")

    @responses.activate
    @mock.patch("cumulusci.core.github.GitHubEnterprise")
    def test_get_github_api_for_repo__enterprise(
        self, GitHubEnterprise, keychain_enterprise
    ):

        gh = get_github_api_for_repo(
            keychain_enterprise, "https://git.enterprise.domain.com/TestOwner/TestRepo/"
        )

        gh.login.assert_called_once_with(token="ATOKEN")

    @responses.activate
    def test_validate_service(self, keychain_enterprise):
        responses.add("GET", "https://api.github.com/user", status=401, headers={})

        with pytest.raises(GithubException):
            validate_service(
                {"username": "BOGUS", "token": "BOGUS"}, keychain_enterprise
            )

    @responses.activate
    def test_validate_gh_enterprise(self, keychain_enterprise):
        keychain_enterprise.set_service(
            "github_enterprise",
            "ent2",
            ServiceConfig(
                {
                    "username": "testusername2",
                    "email": "test2@domain.com",
                    "token": "ATOKEN2",
                    "server_domain": "git.enterprise.domain.com",
                }
            ),
        )

        with pytest.raises(
            GithubException,
            match="More than one Github Enterprise service configured for domain git.enterprise.domain.com",
        ):
            validate_gh_enterprise("git.enterprise.domain.com", keychain_enterprise)

    @responses.activate
    def test_get_auth_from_service(self, keychain_enterprise):
        # github service, should be ignored
        keychain_enterprise.set_service(
            "github",
            "ent",
            ServiceConfig(
                {
                    "username": "testusername",
                    "email": "test@domain.com",
                    "token": "ATOKEN",
                }
            ),
        )
        assert (
            get_auth_from_service("git.enterprise.domain.com", keychain_enterprise)
            == "ATOKEN"
        )

        with pytest.raises(
            ServiceNotConfigured,
            match="No Github Enterprise service configured for domain garbage",
        ):
            get_auth_from_service("garbage", keychain_enterprise)

    @pytest.mark.parametrize(
        "domain,client",
        [
            (None, GitHub),
            ("github.com", GitHub),
            ("api.github.com", GitHub),
            ("git.enterprise.domain.com", GitHubEnterprise),
        ],
    )
    def test_determine_github_client(self, domain, client):
        client_result = _determine_github_client(domain, {})
        assert type(client_result) is client

    @responses.activate
    def test_get_pull_requests_by_head(self, mock_util, repo):
        self.init_github()
        mock_util.mock_pulls(
            pulls=self._get_expected_pull_requests(1),
            head=repo.owner.login + ":" + "some-other-branch",
        )
        pull_requests = get_pull_requests_by_head(repo, "some-other-branch")
        assert 1 == len(pull_requests)

        # ConnectionError present when we reachout with
        # a branch name (url parameter) that we aren't expecting
        with pytest.raises(ConnectionError):
            get_pull_requests_by_head(repo, "does-not-exist")

    @responses.activate
    def test_get_pull_requests_by_head__no_pulls(self, mock_util, repo):
        self.init_github()
        mock_util.mock_pulls()
        pull_requests = get_pull_requests_by_head(repo, "test_branch")
        assert pull_requests == []

        pull_requests = get_pull_requests_by_head(repo, "main")
        assert pull_requests is None

    @responses.activate
    def test_get_pull_request_by_head__multiple_pulls(self, mock_util, repo):
        self.init_github()
        mock_util.mock_pulls(pulls=self._get_expected_pull_requests(2))
        pull_requests = get_pull_requests_by_head(repo, "test_branch")
        assert 2 == len(pull_requests)

    @responses.activate
    def test_get_pull_requests_with_base_branch(self, mock_util, repo):
        self.init_github()
        mock_util.mock_pulls(base="main", head="TestOwner:some-branch")
        pull_requests = get_pull_requests_with_base_branch(
            repo, "main", head="some-branch"
        )
        assert 0 == len(pull_requests)

        responses.reset()
        mock_util.mock_pulls(pulls=self._get_expected_pull_requests(3), base="main")
        pull_requests = get_pull_requests_with_base_branch(repo, "main")
        assert 3 == len(pull_requests)

    @responses.activate
    def test_create_pull_request(self, mock_util, repo, gh_api):
        self.init_github()
        mock_util.mock_pulls(
            method=responses.POST,
            pulls=self._get_expected_pull_request(1, 1, "Test Body"),
        )
        pull_request = create_pull_request(repo, "test-branch")
        assert pull_request is not None
        assert pull_request.body == "Test Body"

    @responses.activate
    def test_is_label_on_pr(self, mock_util, repo, gh_api):
        self.init_github()
        mock_util.add_issue_response(self._get_expected_issue(1))
        mock_util.add_issue_response(self._get_expected_issue(2))
        mock_util.mock_issue_labels(1, responses.GET, ["Octocat", "bogus 1", "bogus 2"])
        mock_util.mock_issue_labels(2, responses.GET, ["bogus 1", "bogus 2"])

        pull_request = ShortPullRequest(self._get_expected_pull_request(1, 1), gh_api)
        pull_request.number = 1
        assert is_label_on_pull_request(repo, pull_request, "Octocat")
        pull_request.number = 2
        assert not is_label_on_pull_request(repo, pull_request, "Octocat")

    @responses.activate
    def test_add_labels_to_pull_request(self, mock_util, repo, gh_api):
        self.init_github()
        mock_util.add_issue_response(self._get_expected_issue(1))
        mock_util.mock_issue_labels(1, method=responses.POST)
        pull_request = ShortPullRequest(self._get_expected_pull_request(1, 1), gh_api)
        pull_request.number = 1

        add_labels_to_pull_request(repo, pull_request, "first", "second", "third")
        body = responses.calls[-1].request.body
        assert "first" in body
        assert "second" in body
        assert "third" in body

    @responses.activate
    def test_get_pull_request_by_commit(self, mock_util, repo, gh_api):
        self.init_github()
        commit_sha = "asdf1234asdf1234"
        mock_util.mock_pull_request_by_commit_sha(commit_sha)
        pull_requests = get_pull_requests_by_commit(gh_api, repo, commit_sha)
        assert len(pull_requests) == 1

    def test_is_pull_request_merged(self, gh_api):
        self.init_github()

        merged_pull_request = ShortPullRequest(
            self._get_expected_pull_request(1, 1), gh_api
        )
        merged_pull_request.merged_at = "DateTimeStr"

        unmerged_pull_request = ShortPullRequest(
            self._get_expected_pull_request(1, 1), gh_api
        )
        unmerged_pull_request.merged_at = None

        assert is_pull_request_merged(merged_pull_request)
        assert not is_pull_request_merged(unmerged_pull_request)

    def test_markdown_link_to_pr(self, gh_api):
        self.init_github()
        pr = ShortPullRequest(self._get_expected_pull_request(1, 1), gh_api)
        actual_link = markdown_link_to_pr(pr)
        expected_link = f"{pr.title} [[PR{pr.number}]({pr.html_url})]"

        assert expected_link == actual_link

    @responses.activate
    def test_create_gist(self, gh_api, mock_util):
        self.init_github()

        description = "Test Gist Creation"
        filename = "error_output.txt"
        content = "Hello there gist!"
        files = {filename: content}

        self.mock_gist(description, files)
        gist = create_gist(gh_api, description, files)

        expected_url = f"https://gist.github.com/{gist.id}"
        assert expected_url == gist.html_url

    def mock_gist(self, description, files):
        responses.add(
            method=responses.POST,
            url="https://api.github.com/gists",
            json=self._get_expected_gist(description, files),
            status=201,
        )

    @responses.activate
    def test_create_gist_no_scope(self, gh_api):
        responses.add(
            method=responses.POST,
            url="https://api.github.com/gists",
            status=403,
        )

        self.init_github()

        description = "Test Gist Creation"
        filename = "error_output.txt"
        content = "Hello there gist!"
        files = {filename: content}

        with pytest.raises(GithubApiError, match="scopes: gist"):
            create_gist(gh_api, description, files)

    @responses.activate
    def test_get_tag_by_name(self, repo):
        self.init_github()
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/tag_SHA",
            json=self._get_expected_tag_ref("tag_SHA", "tag_SHA"),
            status=200,
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/tags/tag_SHA",
            json=self._get_expected_tag("beta/1.0", "tag_SHA"),
            status=200,
        )
        tag = get_tag_by_name(repo, "tag_SHA")
        assert tag.tag == "beta/1.0"

    @responses.activate
    def test_get_tag_by_name__404(self, repo):
        self.init_github()
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/tag_SHA",
            json=self._get_expected_tag_ref("tag_SHA", "tag_SHA"),
            status=200,
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/tags/tag_SHA",
            json=self._get_expected_tag("beta/1.0", "tag_SHA"),
            status=404,
        )
        with pytest.raises(GithubApiNotFoundError):
            get_tag_by_name(repo, "tag_SHA")

    @responses.activate
    def test_current_tag_is_lightweight(self, repo):
        self.init_github()
        light_tag = self._get_expected_tag_ref("tag_SHA", "tag_SHA")
        light_tag["object"]["type"] = "commit"
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/tag_SHA",
            json=light_tag,
            status=200,
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/tags/tag_SHA",
            json=self._get_expected_tag("beta/1.0", "tag_SHA"),
            status=404,
        )
        with pytest.raises(GithubApiNotFoundError) as exc:
            get_tag_by_name(repo, "tag_SHA")

        assert "not an annotated tag" in str(exc)

    @responses.activate
    def test_get_ref_by_name(self, repo):
        self.init_github()
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/tag_SHA",
            json=self._get_expected_tag_ref("tag_SHA", "tag_SHA"),
            status=200,
        )
        ref = get_ref_for_tag(repo, "tag_SHA")
        assert ref.object.sha == "tag_SHA"

    @responses.activate
    def test_get_ref_by_name__404(self, repo):
        self.init_github()
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/tag_SHA",
            json=self._get_expected_tag_ref("tag_SHA", "tag_SHA"),
            status=404,
        )
        with pytest.raises(GithubApiNotFoundError):
            get_ref_for_tag(repo, "tag_SHA")

    @responses.activate
    def test_get_version_id_from_tag(self, repo):
        self.init_github()
        responses.add(  # the ref for the tag is fetched first
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/test-tag-name",
            json=self._get_expected_tag_ref("test-tag-name", "tag_SHA"),
            status=200,
        )
        tag_message = """Release of Test Package\nversion_id: 04t000000000000\n\ndependencies: []"""
        responses.add(  # then we fetch that actual tag with the ref
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/tags/tag_SHA",
            json=self._get_expected_tag("beta/1.0", "tag_SHA", message=tag_message),
            status=200,
        )
        version_id = get_version_id_from_tag(repo, "test-tag-name")
        assert version_id == "04t000000000000"

    @responses.activate
    def test_get_version_id_from_tag__dependency_error(self, repo):
        self.init_github()
        responses.add(  # the ref for the tag is fetched first
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/test-tag-name",
            json=self._get_expected_tag_ref("test-tag-name", "tag_SHA"),
            status=200,
        )
        tag_message = (
            """Release of Test Package\nversion_id: invalid_id\n\ndependencies: []"""
        )
        responses.add(  # then we fetch that actual tag with the ref
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/tags/tag_SHA",
            json=self._get_expected_tag("beta/1.0", "tag_SHA", message=tag_message),
            status=200,
        )
        with pytest.raises(DependencyLookupError):
            get_version_id_from_tag(repo, "test-tag-name")

    @responses.activate
    def test_get_commit(self, repo):
        self.init_github()
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/commits/DUMMY_SHA",
            json=self._get_expected_commit("DUMMY_SHA"),
            status=200,
        )
        commit = get_commit(repo, "DUMMY_SHA")
        assert commit.sha == "DUMMY_SHA"

    @responses.activate
    def test_get_commit__dependency_error(self, repo):
        self.init_github()
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/commits/DUMMY_SHA",
            json={"message": "Could not verify object"},
            status=422,
        )
        with pytest.raises(
            DependencyLookupError, match="Could not find commit DUMMY_SHA on GitHub"
        ):
            get_commit(repo, "DUMMY_SHA")

    def test_get_oauth_scopes(self):
        resp = Response()
        resp.headers["X-OAuth-Scopes"] = "repo, user"
        assert {"repo", "user"} == get_oauth_scopes(resp)

        resp = Response()
        resp.headers["X-OAuth-Scopes"] = "repo"
        assert {"repo"} == get_oauth_scopes(resp)

        resp = Response()
        assert set() == get_oauth_scopes(resp)

    def test_get_sso_disabled_orgs(self):
        resp = Response()
        assert [] == get_sso_disabled_orgs(resp)

        resp = Response()
        resp.headers["X-Github-Sso"] = "partial-results; organizations=0810298,20348880"
        assert ["0810298", "20348880"] == get_sso_disabled_orgs(resp)

    def test_check_github_sso_no_forbidden(self):
        resp = Response()
        resp.status_code = 401
        exc = AuthenticationFailed(resp)
        assert check_github_sso_auth(exc) == ""

        resp.status_code = 403
        exc = ForbiddenError(resp)
        assert check_github_sso_auth(exc) == ""

    @mock.patch("webbrowser.open")
    def test_check_github_sso_unauthorized_token(self, browser_open):
        resp = Response()
        resp.status_code = 403
        auth_url = "https://github.com/orgs/foo/sso?authorization_request=longhash"
        resp.headers["X-Github-Sso"] = f"required; url={auth_url}"
        exc = ForbiddenError(resp)

        check_github_sso_auth(exc)

        browser_open.assert_called_with(auth_url)

    def test_check_github_sso_partial_auth(self):
        resp = Response()
        resp.status_code = 403
        resp.headers["X-Github-Sso"] = "partial-results; organizations=0810298,20348880"
        exc = ForbiddenError(resp)

        expected_err_msg = f"{SSO_WARNING} ['0810298', '20348880']"
        actual_error_msg = check_github_sso_auth(exc).strip()
        assert expected_err_msg == actual_error_msg

    def test_github_oauth_org_restricted(self):
        resp = Response()
        resp.status_code = 403
        body = {"message": "organization has enabled OAuth App access restriction"}
        resp._content = json.dumps(body)
        exc = ForbiddenError(resp)

        expected_warning = "You may also use a Personal Access Token as a workaround."
        actual_error_msg = warn_oauth_restricted(exc)
        assert expected_warning in actual_error_msg

    def test_format_gh3_exc_retry(self):
        resp = Response()
        resp.status_code = 401
        message = "Max retries exceeded with url: foo (Caused by ResponseError('too many 401 error responses',))"
        base_exc = RetryError(message, response=resp)
        exc = TransportError(base_exc)
        assert UNAUTHORIZED_WARNING == format_github3_exception(exc)

    def test_format_gh3_self_signed_ssl(self):
        resp = Response()
        resp.status_code = 401
        message = "[SSL: CERTIFICATE_VERIFY_FAILED] certificate verify failed: self signed certificate in certificate chain"
        base_exc = SSLError(message, response=resp)
        exc = ConnectionError(base_exc)
        assert SELF_SIGNED_WARNING == format_github3_exception(exc)

        # Test passsthrough of other conenction errors
        base_exc = SSLError(response=resp)
        exc = ConnectionError(base_exc)
        assert format_github3_exception(exc) == ""

    def test_format_url_from_exc(self):
        resp = Response()
        resp.status_code = 401
        resp.url = expected_url = "http://example.com"
        message = "Max retries exceeded with url: foo (Caused by ResponseError('too many 401 error responses',))"
        base_exc = RetryError(message, response=resp)
        exc = TransportError(base_exc)
        resp_exc = ResponseError(resp)
        assert expected_url == request_url_from_exc(exc)
        assert expected_url == request_url_from_exc(resp_exc)

    def test_catch_common_decorator(self):
        resp = Response()
        resp.status_code = 403
        resp.headers["X-Github-Sso"] = "partial-results; organizations=0810298,20348880"
        resp.url = "http://zombo.com"

        expected_err_msg = "http://zombo.com\nResults may be incomplete. You have not granted your Personal Access token access to the following organizations: ['0810298', '20348880']"

        @catch_common_github_auth_errors
        def test_func():
            raise ForbiddenError(resp)

        with pytest.raises(GithubApiError) as exc:
            test_func()
            actual_error_msg = exc.message
            assert expected_err_msg == actual_error_msg

    @responses.activate
    def test_catch_common_decorator_transport_url(self):
        mock_rsp = responses.Response(
            method=responses.GET,
            url="https://api.github.com/rate_limit",
            status=401,
            headers={"X-Github-Sso": "partial-results; organizations=0810298,20348880"},
        )
        responses.add(mock_rsp)

        rsp = requests.get("https://api.github.com/rate_limit")
        responses.calls.reset()

        retry_err = RetryError("too many 401 error responses")
        retry_err.response = rsp
        mock_rsp.body = retry_err

        @catch_common_github_auth_errors
        def test_func():
            GitHub().rate_limit()

        with pytest.raises(
            GithubApiError, match="^https://api.github.com/rate_limit\n"
        ):
            test_func()

    def test_catch_common_decorator_ignores(self):
        resp = Response()
        resp.status_code = 401

        @catch_common_github_auth_errors
        def test_func():
            e = RequestException(response=resp)
            raise TransportError(e)

        with pytest.raises(TransportError):
            test_func()

    @responses.activate
    def test_catch_common_decorator_connection_error(self):

        mock_rsp = responses.Response(
            method=responses.GET,
            url="https://api.github.com/rate_limit",
            body=ConnectionError("self signed certificate"),
        )
        responses.add(mock_rsp)

        @catch_common_github_auth_errors
        def test_func():
            GitHub().rate_limit()

        with pytest.raises(
            GithubApiError,
            match="(self-signed certificate in the certificate chain)",
        ):
            test_func()

    @responses.activate
    def test_validate_no_repo_exc(self, keychain_enterprise):
        service_dict = {
            "username": "e2ac67",
            "token": "ghp_cf83e1357eefb8bdf1542850d66d8007d620e4",
            "email": "testerson@test.com",
        }
        responses.add(
            responses.GET,
            "https://api.github.com/user",
            json={
                "login": "e2ac67",
                "id": 91303375,
                "node_id": "MDQ6VXNlcjkxMzAzMzc1",
                "avatar_url": "https://avatars.githubusercontent.com/u/91303375?v=4",
                "gravatar_id": "",
                "url": "https://api.github.com/users/e2ac67",
                "html_url": "https://github.com/e2ac67",
                "followers_url": "https://api.github.com/users/e2ac67/followers",
                "following_url": "https://api.github.com/users/e2ac67/following{/other_user}",
                "gists_url": "https://api.github.com/users/e2ac67/gists{/gist_id}",
                "starred_url": "https://api.github.com/users/e2ac67/starred{/owner}{/repo}",
                "subscriptions_url": "https://api.github.com/users/e2ac67/subscriptions",
                "organizations_url": "https://api.github.com/users/e2ac67/orgs",
                "repos_url": "https://api.github.com/users/e2ac67/repos",
                "events_url": "https://api.github.com/users/e2ac67/events{/privacy}",
                "received_events_url": "https://api.github.com/users/e2ac67/received_events",
                "type": "User",
                "site_admin": False,
                "name": None,
                "company": None,
                "blog": "",
                "location": None,
                "email": None,
                "hireable": None,
                "bio": None,
                "twitter_username": None,
                "public_repos": 0,
                "public_gists": 0,
                "followers": 0,
                "following": 0,
                "created_at": "2021-09-24T03:53:02Z",
                "updated_at": "2021-09-24T03:59:40Z",
            },
        )
        responses.add(
            responses.GET,
            "https://api.github.com/user/orgs",
            json=[],
        )
        responses.add(
            responses.GET,
            "https://api.github.com/user/repos",
            json=[],
            headers={
                "GitHub-Authentication-Token-Expiration": "2021-10-07 19:07:53 UTC",
                "X-OAuth-Scopes": "gist, repo",
            },
        )
        updated_dict = validate_service(service_dict, keychain_enterprise)
        expected_dict = {
            "username": "e2ac67",
            "token": "ghp_cf83e1357eefb8bdf1542850d66d8007d620e4",
            "email": "testerson@test.com",
            "Organizations": "",
            "scopes": "gist, repo",
            "expires": "2021-10-07 19:07:53 UTC",
        }
        assert expected_dict == updated_dict

    @responses.activate
    def test_validate_bad_auth(self, keychain_enterprise):
        service_dict = {
            "username": "e2ac67",
            "token": "bad_cf83e1357eefb8bdf1542850d66d8007d620e4",
            "email": "testerson@test.com",
        }

        responses.add(
            responses.GET,
            "https://api.github.com/user",
            json={
                "message": "Bad credentials",
                "documentation_url": "https://docs.github.com/rest",
            },
            status=401,
        )

        with pytest.raises(cumulusci.core.exceptions.GithubException) as e:
            validate_service(service_dict, keychain_enterprise)
        assert "401" in str(e.value)

    @mock.patch("webbrowser.open")
    @mock.patch("cumulusci.core.github.get_device_code")
    @mock.patch("cumulusci.core.github.get_device_oauth_token", autospec=True)
    def test_get_oauth_device_flow_token(
        self,
        get_token,
        get_code,
        browser_open,
    ):
        device_config = {
            "device_code": "36482450e39b7f27d9a145a96898d29365a4e73f",
            "user_code": "3E15-9D06",
            "verification_uri": "https://github.com/login/device",
            "expires_in": 899,
            "interval": 5,
        }
        get_code.return_value = device_config
        get_token.return_value = {
            "access_token": "expected_access_token",
            "token_type": "bearer",
            "scope": "gist,repo",
        }

        returned_token = get_oauth_device_flow_token()

        assert returned_token == "expected_access_token"
        get_token.assert_called_once()
        get_code.assert_called_once()
        browser_open.assert_called_with("https://github.com/login/device")

    @responses.activate
    @pytest.mark.parametrize(
        ("base_url", "endpoint"),
        (
            ("https://api.github.com", "https://api.github.com/graphql"),
            (
                "https://github.enterprise.server/api/v3",
                "https://github.enterprise.server/api/graphql",
            ),
        ),
    )
    def test_get_latest_prerelease(self, base_url, endpoint):
        expected_tag = "beta/1.0-Beta_1"
        query_result = self._get_expected_prerelease_tag_gql(expected_tag)

        responses.add(
            "POST",
            endpoint,
            json=query_result,
        )

        repo: Repository = mock.MagicMock()
        repo.session = mock.MagicMock()
        repo.session.base_url = base_url
        repo.session.request = requests.request

        get_latest_prerelease(repo=repo)
        assert responses.assert_call_count(endpoint, 1)
        repo.release_from_tag.assert_called_once_with(expected_tag)



"""Salesforce uses a very specialized subset of XML for metadata.
This module has a parser and tree model for dealing with it in
a very Pythonic way.

For example:

>>> from cumulusci.utils.xml.metadata_tree import parse
>>> Recipe = parse("recipe.xml")
>>> print(Recipe.ingredients[0].quantity)

Or to give a more Salesforce-y example:

>>> Package = parse("./cumulusci/files/admin_profile.xml")
>>> print(Package.types[0].members[1].text)
Account
"""

from typing import Generator, Union

from lxml import etree

from . import lxml_parse_file, lxml_parse_string
from .salesforce_encoding import serialize_xml_for_salesforce

METADATA_NAMESPACE = "http://soap.sforce.com/2006/04/metadata"


def parse(source):
    """Parse a file by path or file object into a Metadata Tree

    The parse() function supports any of the following sources:

        * an open file object (make sure to open it in binary mode)
        * a file-like object that has a .read(byte_count) method returning a byte string on each call
        * a filename string or pathlib Path
        * an HTTP or FTP URL string

    """
    if hasattr(source, "open"):  # for pathlib.Path objects
        with source.open(encoding="utf-8") as stream:
            doc = lxml_parse_file(stream)
    else:
        doc = lxml_parse_file(source)
    return MetadataElement(doc.getroot())


def fromstring(source):
    """Parse a Metadata Tree from a string"""
    return MetadataElement(lxml_parse_string(source).getroot())


def parse_package_xml_types(feildName, source_xml_tree):
    """ "Parse metadata types based on the  feildName and map based on the type"""
    xml_map = {}
    for type in source_xml_tree.types:
        members = []
        try:
            for member in type.members:
                members.append(member.text)
        except AttributeError:  # Exception if there are no members for a type
            pass
        xml_map[type[feildName].text] = members
    return xml_map


class MetadataElement:
    '''A class for representing Metadata in a Pythonic tree.

    After you parse into a MetadataElement tree, you can refer
    to child elements like this:

    >>> from cumulusci.utils.xml import metadata_tree
    >>> Package = metadata_tree.fromstring(b"""<?xml version="1.0" encoding="UTF-8"?>
    ... <Package xmlns="http://soap.sforce.com/2006/04/metadata">
    ...     <types>
    ...         <members>CaseReason</members>
    ...         <members>CaseStatus</members>
    ...         <members>CaseType</members>
    ...         <name>StandardValueSet</name>
    ...     </types>
    ...     <version>46.0</version>
    ... </Package>""")
    >>> Package.types.name.text
    'StandardValueSet'

    Or you can refer to them like this:

    >>> Package["types"]["name"].text
    'StandardValueSet'

    That might be convenient if you had the key names in a variable or if you had to refer to an element name
    which would clash with a Python keyword (like `import`) or an instance method (like `text`).

    You can refer to members of lists by index:

    >>> Package["types"]["members"][1]
    <members>CaseStatus</members> element

    There are also methods for finding, appending, inserting and removing nodes, which have their own documentation.
    '''

    __slots__ = ["_element", "_parent", "_ns", "tag"]

    def __init__(self, element: etree._Element, parent: etree._Element = None):
        assert isinstance(element, etree._Element)
        self._element = element
        self._parent = parent
        self._ns = next(iter(element.nsmap.values()))
        self.tag = element.tag.split("}")[1]

    @property
    def text(self):
        if len(self._element):  # if self has any element children
            return self._get_child("text")
        return self._element.text

    @text.setter
    def text(self, text):
        self._element.text = text

    def _wrap_element(self, child: etree._Element):
        return MetadataElement(child, self._element)

    def _add_namespace(self, tag):
        return "{%s}%s" % (self._ns, tag)

    def _get_child(self, childname):
        child_element = self._element.find(self._add_namespace(childname))
        if child_element is None:
            raise AttributeError(f"{childname} not found in {self.tag}")
        return self._wrap_element(child_element)

    def _create_child(self, tag, text=None):
        element = etree.Element(self._add_namespace(tag))
        element.text = text
        return self._wrap_element(element)

    def __getattr__(self, childname):
        return self._get_child(childname)

    def __getitem__(self, item: Union[int, str]):
        """You can get either a child-element by name or a sibling element by index.

        The "sibling element" part may seem non-intuitive but it works like this:

        >>> Recipe = parse("recipe.xml")
        >>> Recipe.ingredients[2]

        First it will evaluate `Recipe.ingredients` and generate an Element for the
        first ingredient.

        Then Python will evaluate the `[2]` and the system will look for the third *sibling*
        of that first ingredient.
        """
        if isinstance(item, int):
            siblings = self._parent.findall(self._element.tag)
            return MetadataElement(siblings[item], self._parent)
        elif isinstance(item, str):
            return self._get_child(item)
        else:
            raise TypeError(
                f"Indices must be integers or strings, not {type(item)}"
            )  # # pragma: no cover

    def append(self, tag: str, text: str = None):
        '''Append a new element at the appropriate place.

        If the parent element (self) already one or more children that match,
        the new element follows the last one.

        Otherwise, the new element goes to the bottom of the parent (self).

        >>> Package = metadata_tree.fromstring(b"""<?xml version="1.0" encoding="UTF-8"?>
        ... <Package xmlns="http://soap.sforce.com/2006/04/metadata">
        ...     <types>
        ...         <members>CaseReason</members>
        ...         <members>CaseStatus</members>
        ...         <members>CaseType</members>
        ...         <name>StandardValueSet</name>
        ...     </types>
        ...     <version>46.0</version>
        ... </Package>""")
        >>> Package.types.append("members", "CaseOfBeer")
        <members>CaseOfBeer</members> element
        >>> print(Package.types.tostring())
        <types xmlns="http://soap.sforce.com/2006/04/metadata">
            <members>CaseReason</members>
            <members>CaseStatus</members>
            <members>CaseType</members>
            <members>CaseOfBeer</members>
            <name>StandardValueSet</name>
        </types>
        '''
        newchild = self._create_child(tag, text)
        same_elements = self._element.findall(self._add_namespace(tag))
        if same_elements:
            last = same_elements[-1]
            index = self._element.index(last)
            self._element.insert(index + 1, newchild._element)
        else:
            self._element.append(newchild._element)
        return newchild

    def insert(self, index: int, tag: str, text: str = None):
        """Insert at a particular index.

        Tag and text can be supplied. Return value is the new element.

        append is preferable because it ensures that nodes are inserted
        in the right "group".

        If you need to get to a particular place in a "group" then insert_before
        and insert_after are preferable.

        If all else fails then you can use this one to precisely insert right
        where you want it but you're responsible for adhering to Salesforce's
        grouping rules.
        """
        newchild = self._create_child(tag, text)
        self._element.insert(index, newchild._element)
        return newchild

    def insert_before(self, oldElement: "MetadataElement", tag: str, text: str = None):
        """Insert before some other element

        Tag and text can be supplied. Return value is the new element."""
        index = self._element.index(oldElement._element)
        return self.insert(index, tag, text)

    def insert_after(self, oldElement: "MetadataElement", tag: str, text: str = None):
        """Insert after some other element

        Tag and text can be supplied. Return value is the new element."""

        index = self._element.index(oldElement._element)
        return self.insert(index + 1, tag, text)

    def remove(self, metadata_element: "MetadataElement") -> None:
        """Remove an element from its parent (self)"""
        self._element.remove(metadata_element._element)

    def find(self, tag, **kwargs):
        """Find a single direct child-elements with name `tag`"""
        return next(self._findall(tag, kwargs), None)

    def findall(self, tag, **kwargs):
        """Find all direct child-elements with name `tag`"""
        return list(self._findall(tag, kwargs))

    def _sub_element_matches_spec(self, e: etree._Element, name: str, value):
        matching_subelement = e.find(self._add_namespace(name))
        if matching_subelement is None and name != "text":
            return value is None
        elif matching_subelement is not None:
            return matching_subelement.text == value
        else:  # matching_subelement is None and name == "text"
            return e.text == value

    def _findall(self, type, kwargs: dict) -> Generator:
        def matches(e):
            return all(
                self._sub_element_matches_spec(e, name, value)
                for name, value in kwargs.items()
            )

        return (
            self._wrap_element(e)
            for e in self._element.findall(self._add_namespace(type))
            if matches(e)
        )

    def tostring(self, xml_declaration=False, include_parent_namespaces=False):
        """Serialize back to XML.

        The XML Declaration is optional and can be controlled by `xml_declaration`"""
        doc = etree.ElementTree(self._element)
        etree.indent(doc, space="    ")
        return serialize_xml_for_salesforce(
            doc,
            xml_declaration=xml_declaration,
            include_parent_namespaces=include_parent_namespaces,
        )

    def __eq__(self, other: "MetadataElement"):
        eq = self._element == other._element
        if eq:
            assert self._parent == other._parent
        return eq

    def __repr__(self):
        children = self._element.getchildren()
        if children:
            contents = f"<!-- {len(children)} children -->"
        elif self.text:
            contents = f"{self.text.strip()}"
        else:
            contents = ""

        return f"<{self.tag}>{contents}</{self.tag}> element"



# import json
import os
from datetime import datetime

# from ssl import SSLCertVerificationError
from unittest import mock

import pytest

# import requests
import responses
from github3 import GitHub, GitHubEnterprise
from github3.exceptions import (  # ResponseError,
    AuthenticationFailed,
    ConnectionError,
    ForbiddenError,
    TransportError,
)

# from github3.pulls import ShortPullRequest
from github3.repos.repo import Repository
from github3.session import AppInstallationTokenAuth
from requests.exceptions import RetryError, SSLError  # RequestException,
from requests.models import Response

# import cumulusci
import cumulusci.vcs.github.service as github
from cumulusci.core.config import BaseProjectConfig, ServiceConfig, UniversalConfig
from cumulusci.core.exceptions import (  # DependencyLookupError,; GithubApiError,; GithubApiNotFoundError,
    GithubException,
    ServiceNotConfigured,
)
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.tasks.github.tests.util_github_api import GithubApiTestMixin
from cumulusci.tasks.release_notes.tests.utils import MockUtil
from cumulusci.vcs.github.service import (  # GitHubService, add_labels_to_pull_request,; catch_common_github_auth_errors,; create_gist,; create_pull_request,; get_commit,; get_github_api,; get_latest_prerelease,; get_oauth_device_flow_token,; get_pull_requests_by_commit,; get_pull_requests_by_head,; get_pull_requests_with_base_branch,; get_ref_for_tag,; get_tag_by_name,; get_version_id_from_tag,; is_label_on_pull_request,; is_pull_request_merged,; markdown_link_to_pr,; request_url_from_exc,; validate_service,; warn_oauth_restricted,
    SELF_SIGNED_WARNING,
    SSO_WARNING,
    UNAUTHORIZED_WARNING,
    GitHubEnterpriseService,
    _determine_github_client,
    check_github_scopes,
    check_github_sso_auth,
    format_github3_exception,
    get_auth_from_service,
    get_github_api_for_repo,
    get_oauth_scopes,
    get_sso_disabled_orgs,
    validate_gh_enterprise,
)


class TestGithub(GithubApiTestMixin):
    @classmethod
    def teardown_method(cls):
        # clear cached repo -> installation mapping
        github.INSTALLATIONS.clear()

    @pytest.fixture
    def mock_util(self):
        return MockUtil("TestOwner", "TestRepo")

    @pytest.fixture
    def repo(self, gh_api):
        repo_json = self._get_expected_repo("TestOwner", "TestRepo")
        return Repository(repo_json, gh_api)

    @pytest.fixture
    def keychain_enterprise(self):
        runtime = mock.Mock()
        runtime.project_config = BaseProjectConfig(UniversalConfig(), config={})
        runtime.keychain = BaseProjectKeychain(runtime.project_config, None)
        runtime.keychain.set_service(
            "github_enterprise",
            "ent",
            ServiceConfig(
                {
                    "username": "testusername",
                    "email": "test@domain.com",
                    "token": "ATOKEN",
                    "server_domain": "git.enterprise.domain.com",
                }
            ),
        )
        return runtime.keychain

    @responses.activate
    @mock.patch("github3.apps.create_token")
    def test_get_github_api_for_repo(self, create_token):
        create_token.return_value = "ATOKEN"
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/installation",
            json={
                "id": 1,
                "access_tokens_url": "",
                "account": "",
                "app_id": "",
                "created_at": "",
                "events": "",
                "html_url": "",
                "permissions": "",
                "repositories_url": "",
                "repository_selection": "",
                "single_file_name": "",
                "target_id": "",
                "target_type": "",
                "updated_at": "",
            },
        )
        responses.add(
            "POST",
            "https://api.github.com/app/installations/1/access_tokens",
            status=201,
            json={"token": "ITOKEN", "expires_at": datetime.now().isoformat()},
        )

        with mock.patch.dict(
            os.environ, {"GITHUB_APP_KEY": "bogus", "GITHUB_APP_ID": "1234"}
        ):
            gh = get_github_api_for_repo(None, "https://github.com/TestOwner/TestRepo/")
            assert isinstance(gh.session.auth, AppInstallationTokenAuth)

    @responses.activate
    @mock.patch("github3.apps.create_token")
    def test_get_github_api_for_repo__not_installed(self, create_token):
        create_token.return_value = "ATOKEN"
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/installation",
            status=404,
        )
        with mock.patch.dict(
            os.environ, {"GITHUB_APP_KEY": "bogus", "GITHUB_APP_ID": "1234"}
        ):
            with pytest.raises(GithubException):
                get_github_api_for_repo(None, "https://github.com/TestOwner/TestRepo/")

    @responses.activate
    @mock.patch("cumulusci.vcs.github.service.GitHub")
    def test_get_github_api_for_repo__token(self, GitHub):
        with mock.patch.dict(os.environ, {"GITHUB_TOKEN": "token"}):
            gh = get_github_api_for_repo(None, "https://github.com/TestOwner/TestRepo/")
        gh.login.assert_called_once_with(token="token")

    @responses.activate
    @mock.patch("cumulusci.vcs.github.service.GitHubEnterprise")
    def test_get_github_api_for_repo__enterprise(
        self, GitHubEnterprise, keychain_enterprise
    ):

        gh = get_github_api_for_repo(
            keychain_enterprise, "https://git.enterprise.domain.com/TestOwner/TestRepo/"
        )

        gh.login.assert_called_once_with(token="ATOKEN")

    @responses.activate
    def test_validate_service(self, keychain_enterprise):
        responses.add("GET", "https://api.github.com/user", status=401, headers={})

        with pytest.raises(GithubException):
            GitHubEnterpriseService.validate_service(
                {"username": "BOGUS", "token": "BOGUS"}, keychain_enterprise
            )

    @responses.activate
    def test_validate_gh_enterprise(self, keychain_enterprise):
        keychain_enterprise.set_service(
            "github_enterprise",
            "ent2",
            ServiceConfig(
                {
                    "username": "testusername2",
                    "email": "test2@domain.com",
                    "token": "ATOKEN2",
                    "server_domain": "git.enterprise.domain.com",
                }
            ),
        )

        with pytest.raises(
            GithubException,
            match="More than one Github Enterprise service configured for domain git.enterprise.domain.com",
        ):
            validate_gh_enterprise("git.enterprise.domain.com", keychain_enterprise)

    @responses.activate
    def test_get_auth_from_service(self, keychain_enterprise):
        # github service, should be ignored
        keychain_enterprise.set_service(
            "github",
            "ent",
            ServiceConfig(
                {
                    "username": "testusername",
                    "email": "test@domain.com",
                    "token": "ATOKEN",
                }
            ),
        )
        assert (
            get_auth_from_service("git.enterprise.domain.com", keychain_enterprise)
            == "ATOKEN"
        )

        with pytest.raises(
            ServiceNotConfigured,
            match="No Github Enterprise service configured for domain garbage",
        ):
            get_auth_from_service("garbage", keychain_enterprise)

    @pytest.mark.parametrize(
        "domain,client",
        [
            (None, GitHub),
            ("github.com", GitHub),
            ("api.github.com", GitHub),
            ("git.enterprise.domain.com", GitHubEnterprise),
        ],
    )
    def test_determine_github_client(self, domain, client):
        client_result = _determine_github_client(domain, {})
        assert isinstance(client_result, client)

    def test_get_sso_disabled_orgs(self):
        resp = Response()
        assert [] == get_sso_disabled_orgs(resp)

        resp = Response()
        resp.headers["X-Github-Sso"] = "partial-results; organizations=0810298,20348880"
        assert ["0810298", "20348880"] == get_sso_disabled_orgs(resp)

    def test_format_gh3_exc_retry(self):
        resp = Response()
        resp.status_code = 401
        message = "Max retries exceeded with url: foo (Caused by ResponseError('too many 401 error responses',))"
        base_exc = RetryError(message, response=resp)
        exc = TransportError(base_exc)
        assert UNAUTHORIZED_WARNING == format_github3_exception(exc)

    def test_format_gh3_self_signed_ssl(self):
        resp = Response()
        resp.status_code = 401
        message = "[SSL: CERTIFICATE_VERIFY_FAILED] certificate verify failed: self signed certificate in certificate chain"
        base_exc = SSLError(message, response=resp)
        exc = ConnectionError(base_exc)
        assert SELF_SIGNED_WARNING == format_github3_exception(exc)

        # Test passsthrough of other conenction errors
        base_exc = SSLError(response=resp)
        exc = ConnectionError(base_exc)
        assert format_github3_exception(exc) == ""

    def test_get_oauth_scopes(self):
        resp = Response()
        resp.headers["X-OAuth-Scopes"] = "repo, user"
        assert {"repo", "user"} == get_oauth_scopes(resp)

        resp = Response()
        resp.headers["X-OAuth-Scopes"] = "repo"
        assert {"repo"} == get_oauth_scopes(resp)

        resp = Response()
        assert set() == get_oauth_scopes(resp)

    def test_check_github_sso_no_forbidden(self):
        resp = Response()
        resp.status_code = 401
        exc = AuthenticationFailed(resp)
        assert check_github_sso_auth(exc) == ""

        resp.status_code = 403
        exc = ForbiddenError(resp)
        assert check_github_sso_auth(exc) == ""

    @mock.patch("webbrowser.open")
    def test_check_github_sso_unauthorized_token(self, browser_open):
        resp = Response()
        resp.status_code = 403
        auth_url = "https://github.com/orgs/foo/sso?authorization_request=longhash"
        resp.headers["X-Github-Sso"] = f"required; url={auth_url}"
        exc = ForbiddenError(resp)

        check_github_sso_auth(exc)

        browser_open.assert_called_with(auth_url)

    def test_check_github_sso_partial_auth(self):
        resp = Response()
        resp.status_code = 403
        resp.headers["X-Github-Sso"] = "partial-results; organizations=0810298,20348880"
        exc = ForbiddenError(resp)

        expected_err_msg = f"{SSO_WARNING} ['0810298', '20348880']"
        actual_error_msg = check_github_sso_auth(exc).strip()
        assert expected_err_msg == actual_error_msg


def test_githubrety_init():
    from cumulusci.vcs.github.service import GitHubRety

    retry = GitHubRety(total=1)
    assert isinstance(retry, GitHubRety)


def test_githubrety_increment_calls_super(monkeypatch):
    from requests.packages.urllib3.util import retry as urllib3_retry

    from cumulusci.vcs.github.service import GitHubRety

    called = {}

    def fake_increment(self, *args, **kwargs):
        called["yes"] = True
        return "super-called"

    monkeypatch.setattr(urllib3_retry.Retry, "increment", fake_increment)
    retry = GitHubRety(total=1)
    result = retry.increment()
    assert called["yes"]
    assert result == "super-called"


def test_githubrety_increment_raises_on_cert_error():
    from cumulusci.vcs.github.service import GitHubRety

    retry = GitHubRety(total=1)

    class DummyError(Exception):
        pass

    error = DummyError("CERTIFICATE_VERIFY_FAILED: something bad")
    with pytest.raises(DummyError):
        retry.increment(error=error)


def test_check_github_scopes_wrong_status_code():
    from github3.exceptions import ResponseError

    resp = Response()
    resp.status_code = 401  # Not 403 or 404
    exc = ResponseError(resp)
    assert check_github_scopes(exc) == ""


def test_check_github_scopes_no_missing_scopes():
    from github3.exceptions import ResponseError

    resp = Response()
    resp.status_code = 403
    resp.headers["X-Accepted-OAuth-Scopes"] = "repo, user"
    resp.headers["X-OAuth-Scopes"] = "repo, user"
    resp.url = "https://api.github.com/repos/foo/bar"
    exc = ResponseError(resp)
    assert check_github_scopes(exc) == ""


def test_check_github_scopes_missing_scopes():
    from github3.exceptions import ResponseError

    resp = Response()
    resp.status_code = 403
    resp.headers["X-Accepted-OAuth-Scopes"] = "repo, user"
    resp.headers["X-OAuth-Scopes"] = "repo"
    resp.url = "https://api.github.com/repos/foo/bar"
    exc = ResponseError(resp)
    result = check_github_scopes(exc)
    assert "Your token may be missing the following scopes: user" in result
    assert "Personal access tokens" in result


def test_check_github_scopes_empty_accepted_scopes():
    from github3.exceptions import ResponseError

    resp = Response()
    resp.status_code = 403
    resp.headers["X-Accepted-OAuth-Scopes"] = ""
    resp.headers["X-OAuth-Scopes"] = "repo"
    resp.url = "https://api.github.com/repos/foo/bar"
    exc = ResponseError(resp)
    assert check_github_scopes(exc) == ""


def test_check_github_scopes_gist_special_case_missing_scope():
    from github3.exceptions import ResponseError

    resp = Response()
    resp.status_code = 404
    resp.headers["X-Accepted-OAuth-Scopes"] = ""
    resp.headers["X-OAuth-Scopes"] = ""
    resp.url = "https://api.github.com/gists"
    exc = ResponseError(resp)
    result = check_github_scopes(exc)
    assert "Your token may be missing the following scopes: gist" in result
    assert "Personal access tokens" in result


def test_check_github_scopes_gist_special_case_has_scope():
    from github3.exceptions import ResponseError

    resp = Response()
    resp.status_code = 404
    resp.headers["X-Accepted-OAuth-Scopes"] = ""
    resp.headers["X-OAuth-Scopes"] = "gist"
    resp.url = "https://api.github.com/gists"
    exc = ResponseError(resp)
    assert check_github_scopes(exc) == ""



import typing as T

from cumulusci.salesforce_api.filterable_objects import NOT_EXTRACTABLE
from cumulusci.salesforce_api.org_schema import Schema

from .synthesize_extract_declarations import (
    SimplifiedExtractDeclaration,
    synthesize_declaration_for_sobject,
)


class SObjDependency(T.NamedTuple):
    table_name_from: str
    table_names_to: T.Union[str, T.Tuple[str, ...]]
    field_name: str
    priority: bool = False


def _calculate_dependencies_for_declarations(
    decls: T.Sequence[SimplifiedExtractDeclaration], schema: Schema
) -> T.Dict[str, T.List[SObjDependency]]:
    """Ensure that required lookups are fulfilled for list of SimplifiedExtractDeclarations

    Do this by adding their referent tables (in full) to the extract.
    Module-internal helper function.
    """
    dependencies = {}
    for decl in decls:
        assert isinstance(decl.sf_object, str)
        new_dependencies = _collect_dependencies_for_sobject(
            decl.sf_object, decl.fields, schema, only_required_fields=False
        )
        dependencies.update(new_dependencies)
    return dependencies


def _collect_dependencies_for_sobject(
    source_sfobject: str,
    fields: T.List[str],
    schema: Schema,
    only_required_fields: bool,
) -> T.Dict[str, T.List[SObjDependency]]:
    """Ensure that required lookups are fulfilled for a single SObject

    Do this by adding its referent tables (in full) to the extract.
    Module-internal helper function.
    """
    dependencies = {}
    for field_name in fields:
        field_info = schema[source_sfobject].fields[field_name]
        if not field_info.createable:  # pragma: no cover
            continue
        references = field_info.referenceTo
        if references:
            # Remove RecordType from references
            if "RecordType" in references:
                references.remove("RecordType")
                if not references:
                    continue

            targets = tuple(
                target for target in references if target not in NOT_EXTRACTABLE
            )
            field_disallowed = not targets or not field_info.createable
            field_allowed = not (only_required_fields or field_disallowed)
            if field_info.requiredOnCreate or field_allowed:
                dependencies.setdefault(source_sfobject, []).append(
                    SObjDependency(
                        source_sfobject,
                        targets,
                        field_name,
                        field_info.requiredOnCreate,
                    )
                )

    return dependencies


def extend_declarations_to_include_referenced_tables(
    decl_list: T.Sequence[SimplifiedExtractDeclaration], schema: Schema
) -> T.Sequence[SimplifiedExtractDeclaration]:
    """Extend the declarations to complete required or requested lookups recursively"""
    decls = {decl.sf_object: decl for decl in decl_list}
    dependencies = _calculate_dependencies_for_declarations(decl_list, schema)
    to_process = list(decls.keys())

    while to_process:
        sf_object = to_process.pop()
        assert isinstance(sf_object, str)
        my_dependencies = dependencies.get(sf_object, ())
        for dep in my_dependencies:
            target_tables = dep.table_names_to
            for target_table in target_tables:
                sobj = schema.get(target_table)
                target_extractable = (
                    target_table not in NOT_EXTRACTABLE and sobj and sobj.extractable
                )
                if target_table not in decls and target_extractable:
                    required_fields = [
                        field.name
                        for field in schema[target_table].fields.values()
                        if field.requiredOnCreate
                    ]
                    decls[target_table] = synthesize_declaration_for_sobject(
                        target_table, required_fields, schema[target_table].fields
                    )

                    new_dependencies = _collect_dependencies_for_sobject(
                        target_table,
                        decls[target_table].fields,
                        schema,
                        only_required_fields=True,
                    )
                    dependencies.update(new_dependencies)
                    to_process.append(target_table)

    return list(decls.values())



import itertools
import typing as T

from snowfakery.salesforce import find_record_type_column

from cumulusci.tasks.bulkdata.mapping_parser import MappingStep
from cumulusci.tasks.bulkdata.step import DataOperationType
from cumulusci.utils.yaml.model_parser import CCIModel

"""A mapping transform is a function with the signature:

def func(mapping_steps: T.List[MappingStep], depmap: "DependencyMap") -> T.List[MappingStep]:

It should be idempotent and a pure function (don't mutate arguments).

Ideally the output should be as similar to the input as possible (i.e.
change what you need to change and leave the rest as it was). In particular,
the order of the input steps should be preserved unless reordering them
is the specific task of the transform.
"""

from cumulusci.tasks.bulkdata.generate_mapping_utils.dependency_map import DependencyMap


def sort_steps(
    mapping_steps: T.List[MappingStep], depmap: DependencyMap
) -> T.List[MappingStep]:
    """Sort mapping declaration steps in dependency order"""
    table_order = depmap.get_dependency_order()

    return sorted(mapping_steps, key=lambda step: table_order.index(step.sf_object))


class _MappingStepSignature(CCIModel):
    index: int
    sf_object: T.Optional[str]
    filters: T.Optional[T.Tuple[str]] = ()
    soql_filter: T.Optional[str] = None  # soql_filter property
    action: DataOperationType
    update_key: T.Union[str, T.Tuple[str, ...]] = ()  # only for upserts


def merge_matching_steps(
    steps: T.Iterable[MappingStep], depmap: DependencyMap
) -> T.List[MappingStep]:
    """Merge mapping declaration steps with similar properties"""
    table_order = {obj.sf_object: idx for idx, obj in enumerate(steps)}
    grouped_steps = itertools.groupby(
        steps,
        lambda step: _MappingStepSignature(
            index=table_order[step.sf_object],
            sf_object=step.sf_object,
            filters=tuple(step.filters) or None,
            soql_filter=step.soql_filter,
            action=step.action,
            update_key=step.update_key,
        ),
    )

    new_steps = [_merge_steps(steps) for _group, steps in grouped_steps]
    return new_steps


def _merge_steps(steps: T.Iterable[MappingStep]) -> MappingStep:
    """Merge a groupp of matching steps together"""
    steps = tuple(steps)
    new_props = steps[0].dict()
    new_props["fields"] = {}
    new_props["lookups"] = {}
    for step in steps:
        new_props["fields"].update(step.fields)
        new_props["lookups"].update(step.lookups)
    return MappingStep(**new_props)


def rename_record_type_fields(
    mapping_steps: T.List[MappingStep], depmap: DependencyMap
) -> T.List[MappingStep]:
    """Rename fields like recordtype, recordtypeid, recordtype_id to RecordTypeId"""

    def doit(mapping_step: MappingStep):
        table_name = mapping_step.sf_object
        record_type_col = find_record_type_column(table_name, mapping_step.fields)
        if record_type_col:
            fields = mapping_step.fields.copy()
            fields["RecordTypeId"] = fields.pop(record_type_col)
            return MappingStep(**{**mapping_step.dict(), "fields": fields})
        else:
            return mapping_step

    return list(map(doit, mapping_steps))


def recategorize_lookups(
    mapping_steps: T.List[MappingStep], depmap: DependencyMap
) -> T.List[MappingStep]:
    """Categorize fields which are actually lookups

    Put them in mapping_step.lookup and remove them from mapping_step.fields"""

    def doit(mapping_step: MappingStep):
        table_name = mapping_step.sf_object
        lookups = mapping_step.lookups.copy() or {}
        lookups.update(
            {
                fieldname: {
                    "table": depmap.target_table_for(table_name, fieldname),
                    "key_field": fieldname,
                }
                for fieldname in mapping_step.fields
                if depmap.target_table_for(table_name, fieldname)
            }
        )

        if lookups:
            fields = [
                field for field in mapping_step.fields if field not in lookups.keys()
            ]
            return MappingStep(
                **{**mapping_step.dict(), "lookups": lookups, "fields": fields}
            )
        else:
            return mapping_step

    return list(map(doit, mapping_steps))


# SnowfakeryPersonAccounts: This will be turned on when Snowfakery is
# integrated with this code.
#
# def change_person_contact_sf_object(
#     mapping_steps: T.List[MappingStep], depmap: DependencyMap
# ) -> T.List[MappingStep]:
#     def doit(mapping_step: MappingStep):

#         if mapping_step.table_name == "PersonContact":
#             mapping_step.sf_object = "Contact"

#     return list(map(doit, mapping_steps))



import os
import time
from pathlib import Path

import yaml

from cumulusci.cli.utils import timestamp_file
from cumulusci.core.exceptions import CumulusCIException, TaskOptionsError
from cumulusci.core.tasks import BaseTask
from cumulusci.core.utils import process_bool_arg, process_list_arg
from cumulusci.utils import download_extract_vcs_from_repo, filter_namelist
from cumulusci.vcs.bootstrap import get_repo_from_url
from cumulusci.vcs.models import AbstractRepo
from cumulusci.vcs.utils import get_ref_from_options


class DownloadExtract(BaseTask):
    task_docs = """A task to download and extract files and folders from a Git repository.
    Will download the repository and extract it to a specified directory.
    """

    task_options = {
        "repo_url": {"description": "The url to the repo", "required": True},
        "target_directory": {
            "description": "The directory to extract the repo contents to."
            "If not set, contents will be extracted to the current directory."
            "If set, it must be a relative path from the project root or an absolute path.",
            "required": True,
        },
        "sub_folder": {
            "description": "The subfolder directory to download from the repo."
        },
        "branch": {
            "description": "The branch to fetch from the repo."
            "if 'ref' or 'tag_name' is not set. The default branch will be set."
        },
        "tag_name": {
            "description": "The name of the tag that should be downloaded."
            "Values of 'latest' and 'latest_beta' are also allowed."
            "Not required if 'branch' is set."
            "Required if 'ref' is not set."
        },
        "ref": {
            "description": "The git reference to download. Takes precedence over 'tag_name' and 'branch'."
            "Required if 'tag_name' or 'branch' is not set."
        },
        "include": {
            "description": "A list of paths from repo root to include. Directories must end with a trailing slash."
        },
        "renames": {
            "description": "A list of paths to rename in the target repo, given as `local:` `target:` pairs."
        },
        "force": {"description": "Force Download files in the repo."},
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        self.options["include"] = process_list_arg(self.options.get("include", None))

        self.options["renames"] = self._process_renames(self.options.get("renames", []))

        self.options["force"] = process_bool_arg(self.options.get("force", False))

    def _init_repo(self):
        self.repo: AbstractRepo = get_repo_from_url(
            self.project_config, self.options["repo_url"]
        )
        self._set_ref()
        self.commit = self.repo.get_ref(self.ref).sha

    def _run_task(self):
        self._set_target_directory()

        if not self._check_latest_commit():
            self.logger.info(
                f"Skipping download, no new changes in {self.options['repo_url']} since last run."
            )
            return

        self.logger.info(
            f"{'Force ' if self.options['force'] else ''}Downloading files from {self.options['repo_url']}."
        )

        target = Path(os.path.join(self.options["target_directory"]))
        if not target.exists():
            target.mkdir(parents=True, exist_ok=True)

        self._download_repo_and_extract(target)
        self._rename_files(target)

        with timestamp_file(self.options["target_directory"]) as f:
            yaml.dump({"commit": self.commit, "timestamp": time.time()}, f)

    def _process_renames(self, renamed_paths):
        """
        For each entry in renames, any renames and store them
        in self.local_to_target_paths.
        """
        if not renamed_paths:
            return {}

        is_list_of_dicts = all(isinstance(pair, dict) for pair in renamed_paths)
        dicts_have_correct_keys = is_list_of_dicts and all(
            {"local", "target"} == pair.keys() for pair in renamed_paths
        )

        ERROR_MSG = (
            "Renamed paths must be a list of dicts with `local:` and `target:` keys."
        )
        if not dicts_have_correct_keys:
            raise TaskOptionsError(ERROR_MSG)

        local_to_target_paths = {}

        for rename in renamed_paths:
            local_path = rename.get("local")
            target_path = rename.get("target")

            if local_path and target_path:
                local_to_target_paths[local_path] = target_path
            else:
                raise TaskOptionsError(ERROR_MSG)

        return local_to_target_paths

    def _set_ref(self):
        if "branch" in self.options:
            branch = self.options.get("branch")
            self.ref = f"heads/{branch}"
            return

        try:
            self.ref = get_ref_from_options(self.project_config, self.options)
        except CumulusCIException:
            # If no ref, tag_name or branch is set, default to the repo's default branch
            self.ref = f"heads/{self.repo.default_branch}"

    def _download_repo_and_extract(self, path):
        zf = download_extract_vcs_from_repo(
            self.repo, subfolder=self.options.get("sub_folder", None), ref=self.commit
        )

        included_members = zf.namelist()
        if self.options["include"]:
            included_members = filter_namelist(
                includes=self.options["include"], namelist=zf.namelist()
            )
        zf.extractall(path=path, members=included_members)
        zf.close()

    def _rename_files(self, zip_dir):
        for local_name, target_name in self.options["renames"].items():
            local_path = Path(zip_dir, local_name)
            if local_path.exists():
                target_path = Path(zip_dir, target_name)
                target_path.parent.mkdir(parents=True, exist_ok=True)
                local_path.replace(target_path)

    def _set_target_directory(self):
        """
        Sets the target directory where the repo contents will be extracted.
        If not set, defaults to the current directory.
        """
        # Check if self.options.get("target_directory") is absolute or relative path.
        target_directory = self.options["target_directory"]
        if not os.path.isabs(target_directory):
            # If it's a relative path, make it absolute based on project root
            target_directory = os.path.join(
                self.project_config.repo_root, target_directory
            )
        self.options["target_directory"] = target_directory

    def _check_latest_commit(self):
        """checks for the latest commit in repo, max once per hour"""
        if self.options["force"]:
            self._init_repo()
            return True

        check = True

        timestamp = 0
        commit = ""
        if os.path.isfile(f"{self.options['target_directory']}/cumulus_timestamp"):
            with timestamp_file(self.options["target_directory"]) as f:
                loaded_data = yaml.safe_load(f)
                timestamp = loaded_data.get("timestamp", 0)
                commit = loaded_data.get("commit", "")

        delta = time.time() - timestamp
        check = delta > 3600

        if not check:
            return False

        self._init_repo()
        if self.commit != commit:
            return True

        return False



import logging
from typing import Optional
from unittest import mock

import pytest
import responses

from cumulusci.core.config import ServiceConfig, TaskConfig
from cumulusci.core.dependencies.dependencies import PackageVersionIdDependency
from cumulusci.core.exceptions import CumulusCIException, TaskOptionsError, VcsException
from cumulusci.tasks.github.tests.util_github_api import GithubApiTestMixin
from cumulusci.tasks.salesforce.promote_package_version import PromotePackageVersion
from cumulusci.tests.util import CURRENT_SF_API_VERSION, create_project_config


@pytest.fixture
def project_config():
    project_config = create_project_config()
    project_config.keychain.set_service(
        "github",
        "alias",
        ServiceConfig(
            {
                "username": "TestUser",
                "token": "TestPass",
                "email": "testuser@testdomain.com",
            },
            "alias",
            project_config.keychain,
        ),
    )
    return project_config


@pytest.fixture
def task(project_config, devhub_config, org_config):
    task = PromotePackageVersion(
        project_config,
        TaskConfig(
            {
                "options": {
                    "version_id": "04t000000000000",
                    "auto_promote": False,
                }
            }
        ),
        org_config,
    )
    with mock.patch(
        "cumulusci.tasks.salesforce.promote_package_version.get_devhub_config",
        return_value=devhub_config,
    ):
        task._init_task()
    return task


class TestPromotePackageVersion(GithubApiTestMixin):
    devhub_base_url = (
        f"https://devhub.my.salesforce.com/services/data/v{CURRENT_SF_API_VERSION}"
    )

    def _mock_target_package_api_calls(self, install_key: Optional[str] = None):
        record = {
            "BuildNumber": 0,
            "Id": "main_package",
            "IsReleased": False,
            "MajorVersion": 1,
            "MinorVersion": 0,
            "PatchVersion": 0,
        }
        if install_key:
            record["InstallKey"] = install_key

        responses.add(  # query for main package's Package2Version info
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [record],
                "done": True,
            },
        )
        responses.add(  # Promote the Package2Version
            "PATCH",
            f"{self.devhub_base_url}/tooling/sobjects/Package2Version/main_package",
        )

    def _mock_dependencies(
        self, total_deps: int, num_2gp: int, num_unpromoted: int
    ) -> None:
        """
        Mock all API calls to represent the dependencies requested in params

        @param total_deps: total number of dependencies to mock
        @param num_2gp: number of 2GP dependencies (all others will be 1GP)
        @param num_unpromoted: of the num_2gp packages, how many are not yet promoted
        """
        spv_ids = [
            {"subscriberPackageVersionId": f"04t00000000000{i + 1}"}
            for i in range(total_deps)
        ]
        responses.add(  # query to find dependency packages
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [{"Dependencies": {"ids": spv_ids}}],
                "done": True,
            },
        )
        # mock 1GP dependencies
        for i in range(total_deps - num_2gp):
            self._mock_dependency(i + 1, is_two_gp=False)

        # mock unpromoted 2GP dependencies
        for i in range(num_unpromoted):
            self._mock_dependency(i + 1, is_two_gp=True)

        # mock promoted 2GP dependencies
        for i in range(num_2gp - num_unpromoted):
            self._mock_dependency(i + 1, is_two_gp=True, is_promoted=True)

    def _mock_dependency(
        self, dependency_num: int, is_two_gp: bool = False, is_promoted: bool = False
    ) -> None:
        """Mock the API calls for a single dependency"""
        responses.add(  # query for SubscriberPackageVersion
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [
                    {
                        "SubscriberPackageId": str(dependency_num),
                        "ReleaseState": "Released" if is_promoted else "Beta",
                    }
                ],
                "done": True,
            },
        )
        responses.add(  # query for SubscriberPackage
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 1,
                "records": [{"Name": f"Dependency_Package_{dependency_num}"}],
                "done": True,
            },
        )

        one_gp_json = {"size": 0, "records": [], "done": True}
        two_gp_json = {
            "size": 1,
            "records": [{"Id": f"dep_{dependency_num}", "IsReleased": False}],
            "done": True,
        }
        responses.add(  # query for Package2Version
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json=(two_gp_json if is_two_gp else one_gp_json),
        )

        if is_two_gp:
            responses.add(
                "PATCH",
                f"{self.devhub_base_url}/tooling/sobjects/Package2Version/dep_{dependency_num}",
            )

    def test_run_task__invalid_version_id(
        self, project_config, devhub_config, org_config
    ):
        with pytest.raises(TaskOptionsError):
            PromotePackageVersion(
                project_config,
                TaskConfig({"options": {"version_id": "0Ho000000000000"}}),
                org_config,
            )

    @responses.activate
    def test_run_task(self, task, devhub_config):
        # 20 dependencies, 10 are 2GP, 5 of those are not yet promoted
        self._mock_dependencies(20, 10, 5)
        self._mock_target_package_api_calls()
        with mock.patch(
            "cumulusci.tasks.salesforce.promote_package_version.get_devhub_config",
            return_value=devhub_config,
        ):
            task()

    @responses.activate
    def test_run_task__install_key(self, task, devhub_config):
        # 20 dependencies, 10 are 2GP, 5 of those are not yet promoted
        task.options["install_key"] = "hunter2"
        self._mock_dependencies(20, 10, 5)
        self._mock_target_package_api_calls(install_key="hunter2")
        with mock.patch(
            "cumulusci.tasks.salesforce.promote_package_version.get_devhub_config",
            return_value=devhub_config,
        ):
            with mock.patch(
                "cumulusci.tasks.salesforce.promote_package_version.get_simple_salesforce_connection",
            ) as get_connection:
                task()
                assert (
                    "hunter2"
                    in get_connection.return_value.query_all.call_args_list[0][0][0]
                )

    @responses.activate
    def test_run_task__no_dependencies(self, task, devhub_config):
        self._mock_dependencies(0, 0, 0)
        self._mock_target_package_api_calls()
        with mock.patch(
            "cumulusci.tasks.salesforce.promote_package_version.get_devhub_config",
            return_value=devhub_config,
        ):
            task()

    @responses.activate
    def test_run_task__promote_dependencies(self, task, devhub_config):
        self._mock_dependencies(2, 1, 1)
        self._mock_target_package_api_calls()
        with mock.patch(
            "cumulusci.tasks.salesforce.promote_package_version.get_devhub_config",
            return_value=devhub_config,
        ):
            task.options["promote_dependencies"] = True
            task()

        assert task.return_values["version_id"] == "04t000000000000"
        assert task.return_values["version_number"] == "1.0.0.0"
        assert task.return_values["dependencies"] == [
            PackageVersionIdDependency(version_id="04t000000000001"),
            PackageVersionIdDependency(version_id="04t000000000002"),
        ]

    @responses.activate
    def test_run_task__all_deps_promoted(self, task, devhub_config):
        self._mock_dependencies(4, 4, 4)
        self._mock_target_package_api_calls()
        with mock.patch(
            "cumulusci.tasks.salesforce.promote_package_version.get_devhub_config",
            return_value=devhub_config,
        ):
            task.options["auto_promote"] = True
            task()

    @responses.activate
    def test_run_task__resolve_version_id(self, task, devhub_config):
        responses.add(  # query for repository
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo",
            json=self._get_expected_repo("TestOwner", "TestRepo"),
            status=200,
        )
        responses.add(  # query for releases (project_config.get_latest_tag)
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/releases?per_page=100",
            json=self._get_expected_releases(
                "TestOwner",
                "TestRepo",
                ["beta/1.113-Beta_1", "whatisthis/0.0.1", "release/0.0.1"],
            ),
            status=200,
        )
        responses.add(  # query for ref to tag
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/beta/1.113-Beta_1",
            json=self._get_expected_tag_ref("tag_SHA", "tag_SHA"),
            status=200,
        )
        responses.add(  # query for tag
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/tags/tag_SHA",
            json=self._get_expected_tag(
                "beta/1.0",
                "tag_SHA",
                message="Release for Beta v1.113\n\nversion_id: 04t000000000000\n\ndependencies: []",
            ),
            status=200,
        )
        self._mock_dependencies(2, 1, 0)
        self._mock_target_package_api_calls()
        with mock.patch(
            "cumulusci.tasks.salesforce.promote_package_version.get_devhub_config",
            return_value=devhub_config,
        ):
            task.options["version_id"] = None
            task()

    @responses.activate
    def test_run_task__resolve_version_id_dependency_error(self, task, devhub_config):
        responses.add(  # query for repository
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo",
            json=self._get_expected_repo("TestOwner", "TestRepo"),
            status=200,
        )
        responses.add(  # query for releases (project_config.get_latest_tag)
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/releases?per_page=100",
            json=self._get_expected_releases(
                "TestOwner",
                "TestRepo",
                ["beta/1.113-Beta_1", "whatisthis/0.0.1", "release/0.0.1"],
            ),
            status=200,
        )
        responses.add(  # query for ref to tag
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/beta/1.113-Beta_1",
            json=self._get_expected_tag_ref("tag_SHA", "tag_SHA"),
            status=200,
        )
        responses.add(  # query for tag
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/tags/tag_SHA",
            json=self._get_expected_tag(
                "beta/1.0",
                "tag_SHA",
                message="Release for Beta v1.113\n\ndependencies: []",
            ),
            status=200,
        )
        self._mock_dependencies(2, 1, 0)
        self._mock_target_package_api_calls()
        with mock.patch(
            "cumulusci.tasks.salesforce.promote_package_version.get_devhub_config",
            return_value=devhub_config,
        ):
            task.options["version_id"] = None
            with pytest.raises(VcsException):
                task()

    @responses.activate
    def test_run_task__resolve_version_id_dependency_error_malformed_version(
        self, task, devhub_config, caplog
    ):
        responses.add(  # query for repository
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo",
            json=self._get_expected_repo("TestOwner", "TestRepo"),
            status=200,
        )
        responses.add(  # query for releases (project_config.get_latest_tag)
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/releases?per_page=100",
            json=self._get_expected_releases(
                "TestOwner",
                "TestRepo",
                ["beta/1.113-Beta_1", "whatisthis/0.0.1", "release/0.0.1"],
            ),
            status=200,
        )
        responses.add(  # query for ref to tag
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/beta/1.113-Beta_1",
            json=self._get_expected_tag_ref("tag_SHA", "tag_SHA"),
            status=200,
        )
        responses.add(  # query for tag
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/tags/tag_SHA",
            json=self._get_expected_tag(
                "beta/1.0",
                "tag_SHA",
                message="Release for Beta v1.113\n\nversion_id: malformedId\n\ndependencies: []",
            ),
            status=200,
        )
        self._mock_dependencies(2, 1, 0)
        self._mock_target_package_api_calls()
        with mock.patch(
            "cumulusci.tasks.salesforce.promote_package_version.get_devhub_config",
            return_value=devhub_config,
        ):
            task.options["version_id"] = None
            with pytest.raises(VcsException):
                task()

    def test_process_one_gp_dependencies(self, task, caplog):
        """Ensure proper logging output"""
        dependencies = [
            {
                "is_2gp": False,
                "name": "Dependency 1",
                "release_state": "Beta",
            },
            {
                "is_2gp": True,
                "name": "Dependency 2",
                "release_state": "Beta",
            },
        ]
        task._process_one_gp_deps(dependencies)
        assert (
            "This package has the following 1GP dependencies:"
            == caplog.records[1].message
        )
        assert "Package Name: Dependency 1" in caplog.records[3].message
        assert "Release State: Beta" in caplog.records[4].message

    def test_process_two_gp_dependencies(self, task, caplog):
        """Ensure proper logging output"""
        dependencies = [
            {"is_2gp": False, "name": "Dependency 1", "release_state": "Beta"},
            {
                "is_2gp": True,
                "name": "Dependency 2",
                "release_state": "Beta",
                "is_promoted": False,
                "version_id": "04t000000000002",
            },
        ]
        with caplog.at_level(logging.INFO):
            task._process_two_gp_deps(dependencies)
        assert caplog.records[1].message == "Total 2GP dependencies: 1"
        assert caplog.records[2].message == "Unpromoted 2GP dependencies: 1"
        assert (
            "This package depends on other packages that have not yet been promoted."
            == caplog.records[4].message
        )
        assert "Package Name: Dependency 2" in caplog.records[8].message

    @responses.activate
    def test_query_Package2Version__malformed_request(self, task):
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json=[{"message": "Object type 'Package2' is not supported"}],
            status=400,
        )
        with pytest.raises(TaskOptionsError):
            task._query_Package2Version("04t000000000000")

    @responses.activate
    def test_query_one_tooling(self, task):
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 2,
                "records": [{"name": "Thing_1"}, {"name": "Thing_2"}],
                "done": True,
            },
        )
        obj = task._query_one_tooling(["name"], "sObjectName")
        assert not isinstance(obj, list)

    @responses.activate
    def test_query_tooling__return_none(self, task):
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 0, "records": [], "done": True},
        )
        result = task._query_tooling(["Id", "name"], "sObjectName")
        assert result is None

    @responses.activate
    def test_query_tooling__return_multiple(self, task):
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={
                "size": 2,
                "records": [{"name": "Thing_1"}, {"name": "Thing_2"}],
                "done": True,
            },
        )
        result = task._query_tooling(["Id", "name"], "sObjectName")
        assert isinstance(result, list)
        assert len(result) == 2

    @responses.activate
    def test_query_tooling__raise_error(self, task):
        responses.add(
            "GET",
            f"{self.devhub_base_url}/tooling/query/",
            json={"size": 0, "records": [], "done": True},
        )
        with pytest.raises(
            CumulusCIException,
            match="No records returned for query: SELECT Id, Field__c FROM sObjectName WHERE Id='12345'",
        ):
            task._query_tooling(
                ["Id", "Field__c"], "sObjectName", "Id='12345'", raise_error=True
            )
